#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import dataclasses
import warnings
from random import choice
from typing import cast, List
from unittest import mock

import pandas as pd
from ax.core.arm import Arm
from ax.core.observation import ObservationFeatures
from ax.core.parameter import (
    ChoiceParameter,
    FixedParameter,
    Parameter,
    ParameterType,
    RangeParameter,
)
from ax.core.parameter_constraint import (
    OrderConstraint,
    ParameterConstraint,
    SumConstraint,
)
from ax.core.parameter_distribution import ParameterDistribution
from ax.core.search_space import (
    HierarchicalSearchSpace,
    RobustSearchSpace,
    RobustSearchSpaceDigest,
    SearchSpace,
    SearchSpaceDigest,
)
from ax.core.types import TParameterization
from ax.exceptions.core import UnsupportedError, UserInputError
from ax.utils.common.constants import Keys
from ax.utils.common.testutils import TestCase
from ax.utils.common.typeutils import checked_cast
from ax.utils.testing.core_stubs import (
    get_hierarchical_search_space,
    get_l2_reg_weight_parameter,
    get_lr_parameter,
    get_model_parameter,
    get_num_boost_rounds_parameter,
    get_parameter_constraint,
)

TOTAL_PARAMS = 6
TUNABLE_PARAMS = 4
RANGE_PARAMS = 3


class SearchSpaceTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.a = RangeParameter(
            name="a", parameter_type=ParameterType.FLOAT, lower=0.5, upper=5.5
        )
        self.b = RangeParameter(
            name="b", parameter_type=ParameterType.INT, lower=2, upper=10
        )
        self.c = ChoiceParameter(
            name="c", parameter_type=ParameterType.STRING, values=["foo", "bar", "baz"]
        )
        self.d = FixedParameter(name="d", parameter_type=ParameterType.BOOL, value=True)
        self.e = ChoiceParameter(
            name="e", parameter_type=ParameterType.FLOAT, values=[0.0, 0.1, 0.2, 0.5]
        )
        self.f = RangeParameter(
            name="f",
            parameter_type=ParameterType.INT,
            lower=2,
            upper=10,
            log_scale=True,
        )
        self.g = RangeParameter(
            name="g", parameter_type=ParameterType.FLOAT, lower=0.0, upper=1.0
        )
        self.parameters: List[Parameter] = [
            self.a,
            self.b,
            self.c,
            self.d,
            self.e,
            self.f,
        ]
        self.ss1 = SearchSpace(parameters=self.parameters)
        self.ss2 = SearchSpace(
            parameters=self.parameters,
            parameter_constraints=[
                OrderConstraint(lower_parameter=self.a, upper_parameter=self.b)
            ],
        )
        self.ss1_repr = (
            "SearchSpace("
            "parameters=["
            "RangeParameter(name='a', parameter_type=FLOAT, range=[0.5, 5.5]), "
            "RangeParameter(name='b', parameter_type=INT, range=[2, 10]), "
            "ChoiceParameter(name='c', parameter_type=STRING, "
            "values=['foo', 'bar', 'baz'], is_ordered=False, sort_values=False), "
            "FixedParameter(name='d', parameter_type=BOOL, value=True), "
            "ChoiceParameter(name='e', parameter_type=FLOAT, "
            "values=[0.0, 0.1, 0.2, 0.5], is_ordered=True, sort_values=True), "
            "RangeParameter(name='f', parameter_type=INT, range=[2, 10], "
            "log_scale=True)], "
            "parameter_constraints=[])"
        )
        self.ss2_repr = (
            "SearchSpace("
            "parameters=["
            "RangeParameter(name='a', parameter_type=FLOAT, range=[0.5, 5.5]), "
            "RangeParameter(name='b', parameter_type=INT, range=[2, 10]), "
            "ChoiceParameter(name='c', parameter_type=STRING, "
            "values=['foo', 'bar', 'baz'], is_ordered=False, sort_values=False), "
            "FixedParameter(name='d', parameter_type=BOOL, value=True), "
            "ChoiceParameter(name='e', parameter_type=FLOAT, "
            "values=[0.0, 0.1, 0.2, 0.5], is_ordered=True, sort_values=True), "
            "RangeParameter(name='f', parameter_type=INT, range=[2, 10], "
            "log_scale=True)], "
            "parameter_constraints=[OrderConstraint(a <= b)])"
        )

    def test_Eq(self) -> None:
        ss2 = SearchSpace(
            parameters=self.parameters,
            parameter_constraints=[
                OrderConstraint(lower_parameter=self.a, upper_parameter=self.b)
            ],
        )
        self.assertEqual(self.ss2, ss2)
        self.assertNotEqual(self.ss1, self.ss2)

    def test_Properties(self) -> None:
        self.assertEqual(len(self.ss1.parameters), TOTAL_PARAMS)
        self.assertTrue("a" in self.ss1.parameters)
        self.assertTrue(len(self.ss1.tunable_parameters), TUNABLE_PARAMS)
        self.assertFalse("d" in self.ss1.tunable_parameters)
        self.assertTrue(len(self.ss1.range_parameters), RANGE_PARAMS)
        self.assertFalse("c" in self.ss1.range_parameters)
        self.assertTrue(len(self.ss1.parameter_constraints) == 0)
        self.assertTrue(len(self.ss2.parameter_constraints) == 1)

    def test_Repr(self) -> None:
        self.assertEqual(str(self.ss2), self.ss2_repr)
        self.assertEqual(str(self.ss1), self.ss1_repr)

    def test_Setter(self) -> None:
        new_c = SumConstraint(
            parameters=[self.a, self.b], is_upper_bound=True, bound=10
        )
        self.ss2.add_parameter_constraints([new_c])
        self.assertEqual(len(self.ss2.parameter_constraints), 2)

        self.ss2.set_parameter_constraints([])
        self.assertEqual(len(self.ss2.parameter_constraints), 0)

        update_p = RangeParameter(
            name="b", parameter_type=ParameterType.INT, lower=10, upper=20
        )
        self.ss2.add_parameter(self.g)
        self.assertEqual(len(self.ss2.parameters), TOTAL_PARAMS + 1)

        self.ss2.update_parameter(update_p)
        # pyre-fixme[16]: `Parameter` has no attribute `lower`.
        self.assertEqual(self.ss2.parameters["b"].lower, 10)

    def test_BadConstruction(self) -> None:
        # Duplicate parameter
        with self.assertRaises(ValueError):
            p1 = self.parameters + [self.parameters[0]]
            SearchSpace(parameters=p1, parameter_constraints=[])

        # Constraint on non-existent parameter
        with self.assertRaises(ValueError):
            SearchSpace(
                parameters=self.parameters,
                parameter_constraints=[
                    OrderConstraint(lower_parameter=self.a, upper_parameter=self.g)
                ],
            )

        # Vanilla Constraint on non-existent parameter
        with self.assertRaises(ValueError):
            SearchSpace(
                parameters=self.parameters,
                parameter_constraints=[
                    ParameterConstraint(constraint_dict={"g": 1}, bound=0)
                ],
            )

        # Constraint on non-numeric parameter
        with self.assertRaises(ValueError):
            SearchSpace(
                parameters=self.parameters,
                parameter_constraints=[
                    OrderConstraint(lower_parameter=self.a, upper_parameter=self.d)
                ],
            )

        # Constraint on choice parameter
        with self.assertRaises(ValueError):
            SearchSpace(
                parameters=self.parameters,
                parameter_constraints=[
                    OrderConstraint(lower_parameter=self.a, upper_parameter=self.e)
                ],
            )

        # Constraint on logscale parameter
        with self.assertRaises(ValueError):
            SearchSpace(
                parameters=self.parameters,
                parameter_constraints=[
                    OrderConstraint(lower_parameter=self.a, upper_parameter=self.f)
                ],
            )

        # Constraint on mismatched parameter
        with self.assertRaises(ValueError):
            wrong_a = self.a.clone()
            wrong_a.update_range(upper=10)
            SearchSpace(
                parameters=self.parameters,
                parameter_constraints=[
                    OrderConstraint(lower_parameter=wrong_a, upper_parameter=self.b)
                ],
            )

    def test_BadSetter(self) -> None:
        new_p = RangeParameter(
            name="b", parameter_type=ParameterType.FLOAT, lower=0.0, upper=1.0
        )

        # Add duplicate parameter
        with self.assertRaises(ValueError):
            self.ss1.add_parameter(new_p)

        # Update parameter to different type
        with self.assertRaises(ValueError):
            self.ss1.update_parameter(new_p)

        # Update non-existent parameter
        new_p = RangeParameter(
            name="g", parameter_type=ParameterType.FLOAT, lower=0.0, upper=1.0
        )
        with self.assertRaises(ValueError):
            self.ss1.update_parameter(new_p)

    def test_CheckMembership(self) -> None:
        p_dict = {"a": 1.0, "b": 5, "c": "foo", "d": True, "e": 0.2, "f": 5}

        # Valid
        # pyre-fixme[6]: For 1st param expected `Dict[str, Union[None, bool, float,
        #  int, str]]` but got `Dict[str, Union[float, str]]`.
        self.assertTrue(self.ss2.check_membership(p_dict))

        # Value out of range
        p_dict["a"] = 20.0
        # pyre-fixme[6]: For 1st param expected `Dict[str, Union[None, bool, float,
        #  int, str]]` but got `Dict[str, Union[float, str]]`.
        self.assertFalse(self.ss2.check_membership(p_dict))
        with self.assertRaises(ValueError):
            # pyre-fixme[6]: For 1st param expected `Dict[str, Union[None, bool,
            #  float, int, str]]` but got `Dict[str, Union[float, str]]`.
            self.ss2.check_membership(p_dict, raise_error=True)

        # Violate constraints
        p_dict["a"] = 5.3
        # pyre-fixme[6]: For 1st param expected `Dict[str, Union[None, bool, float,
        #  int, str]]` but got `Dict[str, Union[float, str]]`.
        self.assertFalse(self.ss2.check_membership(p_dict))
        with self.assertRaises(ValueError):
            # pyre-fixme[6]: For 1st param expected `Dict[str, Union[None, bool,
            #  float, int, str]]` but got `Dict[str, Union[float, str]]`.
            self.ss2.check_membership(p_dict, raise_error=True)

        # Incomplete param dict
        p_dict.pop("a")
        # pyre-fixme[6]: For 1st param expected `Dict[str, Union[None, bool, float,
        #  int, str]]` but got `Dict[str, Union[float, str]]`.
        self.assertFalse(self.ss2.check_membership(p_dict))
        with self.assertRaises(ValueError):
            # pyre-fixme[6]: For 1st param expected `Dict[str, Union[None, bool,
            #  float, int, str]]` but got `Dict[str, Union[float, str]]`.
            self.ss2.check_membership(p_dict, raise_error=True)

        # Unknown parameter
        p_dict["q"] = 40
        # pyre-fixme[6]: For 1st param expected `Dict[str, Union[None, bool, float,
        #  int, str]]` but got `Dict[str, Union[float, str]]`.
        self.assertFalse(self.ss2.check_membership(p_dict))
        with self.assertRaises(ValueError):
            # pyre-fixme[6]: For 1st param expected `Dict[str, Union[None, bool,
            #  float, int, str]]` but got `Dict[str, Union[float, str]]`.
            self.ss2.check_membership(p_dict, raise_error=True)

    def test_CheckTypes(self) -> None:
        p_dict = {"a": 1.0, "b": 5, "c": "foo", "d": True, "e": 0.2, "f": 5}

        # Valid
        # pyre-fixme[6]: For 1st param expected `Dict[str, Union[None, bool, float,
        #  int, str]]` but got `Dict[str, Union[float, str]]`.
        self.assertTrue(self.ss2.check_types(p_dict))

        # Invalid type
        p_dict["b"] = 5.2
        # pyre-fixme[6]: For 1st param expected `Dict[str, Union[None, bool, float,
        #  int, str]]` but got `Dict[str, Union[float, str]]`.
        self.assertFalse(self.ss2.check_types(p_dict))
        with self.assertRaises(ValueError):
            # pyre-fixme[6]: For 1st param expected `Dict[str, Union[None, bool,
            #  float, int, str]]` but got `Dict[str, Union[float, str]]`.
            self.ss2.check_types(p_dict, raise_error=True)
        p_dict["b"] = 5

        # Unknown parameter
        p_dict["q"] = 40
        # pyre-fixme[6]: For 1st param expected `Dict[str, Union[None, bool, float,
        #  int, str]]` but got `Dict[str, Union[float, str]]`.
        self.assertFalse(self.ss2.check_types(p_dict))
        with self.assertRaises(ValueError):
            # pyre-fixme[6]: For 1st param expected `Dict[str, Union[None, bool,
            #  float, int, str]]` but got `Dict[str, Union[float, str]]`.
            self.ss2.check_types(p_dict, raise_error=True)

    def test_CastArm(self) -> None:
        p_dict = {"a": 1.0, "b": 5.0, "c": "foo", "d": True, "e": 0.2, "f": 5}

        # Check "b" parameter goes from float to int
        self.assertTrue(isinstance(p_dict["b"], float))
        # pyre-fixme[6]: For 1st param expected `Dict[str, Union[None, bool, float,
        #  int, str]]` but got `Dict[str, Union[float, str]]`.
        new_arm = self.ss2.cast_arm(Arm(p_dict))
        self.assertTrue(isinstance(new_arm.parameters["b"], int))

        # Unknown parameter should be unchanged
        p_dict["q"] = 40
        # pyre-fixme[6]: For 1st param expected `Dict[str, Union[None, bool, float,
        #  int, str]]` but got `Dict[str, Union[float, str]]`.
        new_arm = self.ss2.cast_arm(Arm(p_dict))
        self.assertTrue(isinstance(new_arm.parameters["q"], int))

    def test_Copy(self) -> None:
        a = RangeParameter("a", ParameterType.FLOAT, 1.0, 5.5)
        b = RangeParameter("b", ParameterType.FLOAT, 2.0, 5.5)
        c = ChoiceParameter("c", ParameterType.INT, [2, 3])
        ss = SearchSpace(
            parameters=[a, b, c],
            parameter_constraints=[
                OrderConstraint(lower_parameter=a, upper_parameter=b)
            ],
        )
        ss_copy = ss.clone()
        self.assertEqual(len(ss_copy.parameters), len(ss_copy.parameters))
        self.assertEqual(
            len(ss_copy.parameter_constraints), len(ss_copy.parameter_constraints)
        )

        ss_copy.add_parameter(FixedParameter("d", ParameterType.STRING, "h"))
        self.assertNotEqual(len(ss_copy.parameters), len(ss.parameters))

    def test_OutOfDesignArm(self) -> None:
        arm1 = self.ss1.out_of_design_arm()
        arm2 = self.ss2.out_of_design_arm()
        arm1_nones = [p is None for p in arm1.parameters.values()]
        self.assertTrue(all(arm1_nones))
        self.assertTrue(arm1 == arm2)

    def test_ConstructArm(self) -> None:
        # Test constructing an arm of default values
        arm = self.ss1.construct_arm(name="test")
        self.assertEqual(arm.name, "test")
        for p_name in self.ss1.parameters.keys():
            self.assertTrue(p_name in arm.parameters)
            self.assertEqual(arm.parameters[p_name], None)

        # Test constructing an arm with a custom value
        arm = self.ss1.construct_arm({"a": 1.0})
        for p_name in self.ss1.parameters.keys():
            self.assertTrue(p_name in arm.parameters)
            if p_name == "a":
                self.assertEqual(arm.parameters[p_name], 1.0)
            else:
                self.assertEqual(arm.parameters[p_name], None)

        # Test constructing an arm with a bad param name
        with self.assertRaises(ValueError):
            self.ss1.construct_arm({"IDONTEXIST_a": 1.0})

        # Test constructing an arm with a bad param name
        with self.assertRaises(ValueError):
            self.ss1.construct_arm({"a": "notafloat"})

    def test_search_space_summary_df(self) -> None:
        min_search_space = SearchSpace(
            parameters=[
                RangeParameter(
                    "x1", parameter_type=ParameterType.INT, lower=0, upper=2
                ),
                RangeParameter(
                    "x2",
                    parameter_type=ParameterType.FLOAT,
                    lower=0.1,
                    upper=10,
                ),
            ]
        )
        max_search_space = SearchSpace(
            parameters=[
                RangeParameter(
                    "x1", parameter_type=ParameterType.INT, lower=0, upper=2
                ),
                RangeParameter(
                    "x2",
                    parameter_type=ParameterType.FLOAT,
                    lower=0.1,
                    upper=10,
                    log_scale=True,
                    is_fidelity=True,
                    target_value=10,
                ),
                FixedParameter("x3", parameter_type=ParameterType.BOOL, value=True),
                ChoiceParameter(
                    "x4",
                    parameter_type=ParameterType.STRING,
                    values=["a", "b", "c"],
                    is_ordered=False,
                    dependents={"a": ["x1", "x2"], "b": ["x4", "x5"], "c": ["x6"]},
                ),
                ChoiceParameter(
                    "x5",
                    parameter_type=ParameterType.STRING,
                    values=["d", "e", "f"],
                    is_ordered=True,
                ),
                RangeParameter(
                    name="x6",
                    parameter_type=ParameterType.FLOAT,
                    lower=0.0,
                    upper=1.0,
                ),
            ]
        )
        df = max_search_space.summary_df
        expected_df = pd.DataFrame(
            data={
                "Name": ["x1", "x2", "x3", "x4", "x5", "x6"],
                "Type": ["Range", "Range", "Fixed", "Choice", "Choice", "Range"],
                "Domain": [
                    "range=[0, 2]",
                    "range=[0.1, 10.0]",
                    "value=True",
                    "values=['a', 'b', 'c']",
                    "values=['d', 'e', 'f']",
                    "range=[0.0, 1.0]",
                ],
                "Datatype": ["int", "float", "bool", "string", "string", "float"],
                "Flags": [
                    "None",
                    "fidelity, log_scale",
                    "None",
                    "unordered, hierarchical, unsorted",
                    "ordered, unsorted",
                    "None",
                ],
                "Target Value": ["None", 10.0, "None", "None", "None", "None"],
                "Dependent Parameters": [
                    "None",
                    "None",
                    "None",
                    {"a": ["x1", "x2"], "b": ["x4", "x5"], "c": ["x6"]},
                    "None",
                    "None",
                ],
            }
        )

        pd.testing.assert_frame_equal(df, expected_df)

        df = min_search_space.summary_df
        expected_df = pd.DataFrame(
            data={
                "Name": ["x1", "x2"],
                "Type": ["Range", "Range"],
                "Domain": ["range=[0, 2]", "range=[0.1, 10.0]"],
                "Datatype": ["int", "float"],
            }
        )
        pd.testing.assert_frame_equal(df, expected_df)


class SearchSpaceDigestTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.kwargs = {
            "feature_names": ["a", "b", "c"],
            "bounds": [(0.0, 1.0), (0, 2), (0, 4)],
            "ordinal_features": [1],
            "categorical_features": [2],
            "discrete_choices": {1: [0, 1, 2], 2: [0, 0.25, 4.0]},
            "task_features": [3],
            "fidelity_features": [0],
            "target_values": {0: 1.0},
            "robust_digest": None,
        }

    def test_SearchSpaceDigest(self) -> None:
        # test required fields
        with self.assertRaises(TypeError):
            # pyre-fixme[20]: Argument `feature_names` expected.
            SearchSpaceDigest(bounds=[])
        with self.assertRaises(TypeError):
            # pyre-fixme[20]: Argument `bounds` expected.
            SearchSpaceDigest(feature_names=[])
        # test instantiation
        ssd = SearchSpaceDigest(**self.kwargs)
        self.assertEqual(dataclasses.asdict(ssd), self.kwargs)
        # test default instatiation
        for arg in self.kwargs:
            if arg in {"feature_names", "bounds"}:
                continue
            ssd = SearchSpaceDigest(
                **{k: v for k, v in self.kwargs.items() if k != arg}
            )


class RobustSearchSpaceDigestTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.kwargs = {
            "sample_param_perturbations": lambda: 1,
            "sample_environmental": lambda: 2,
            "environmental_variables": ["a"],
            "multiplicative": False,
        }

    def test_robust_search_space_digest(self) -> None:
        # test post init
        with self.assertRaises(UserInputError):
            RobustSearchSpaceDigest()
        # test instantiation
        rssd = RobustSearchSpaceDigest(**self.kwargs)
        self.assertEqual(dataclasses.asdict(rssd), self.kwargs)
        # test default instantiation
        for arg in self.kwargs:
            rssd = RobustSearchSpaceDigest(
                **{k: v for k, v in self.kwargs.items() if k != arg}
            )


class HierarchicalSearchSpaceTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.model_parameter = get_model_parameter()
        self.lr_parameter = get_lr_parameter()
        self.l2_reg_weight_parameter = get_l2_reg_weight_parameter()
        self.num_boost_rounds_parameter = get_num_boost_rounds_parameter()
        self.hss_1 = get_hierarchical_search_space()
        self.use_linear_parameter = ChoiceParameter(
            name="use_linear",  # Contrived!
            parameter_type=ParameterType.BOOL,
            values=[True, False],
            dependents={
                True: ["model"],
            },
        )
        self.hss_2 = HierarchicalSearchSpace(
            parameters=[
                self.use_linear_parameter,
                self.model_parameter,
                self.lr_parameter,
                self.l2_reg_weight_parameter,
                self.num_boost_rounds_parameter,
            ]
        )
        self.hss_with_fixed = HierarchicalSearchSpace(
            parameters=[
                self.use_linear_parameter,
                FixedParameter(
                    name="model",
                    value="fixed_model",
                    parameter_type=ParameterType.STRING,
                ),
            ]
        )
        self.model_2_parameter = ChoiceParameter(
            name="model_2",
            parameter_type=ParameterType.STRING,
            values=["Linear", "XGBoost"],
            dependents={
                "Linear": ["learning_rate", "l2_reg_weight"],
                "XGBoost": ["num_boost_rounds"],
            },
        )
        self.hss_with_constraints = HierarchicalSearchSpace(
            parameters=[
                self.model_parameter,
                self.lr_parameter,
                self.l2_reg_weight_parameter,
                self.num_boost_rounds_parameter,
            ],
            parameter_constraints=[
                get_parameter_constraint(
                    param_x=self.lr_parameter.name,
                    param_y=self.l2_reg_weight_parameter.name,
                ),
            ],
        )
        self.hss_1_arm_1_flat = Arm(
            parameters={
                "model": "Linear",
                "learning_rate": 0.01,
                "l2_reg_weight": 0.0001,
                "num_boost_rounds": 12,
            }
        )
        self.hss_1_arm_2_flat = Arm(
            parameters={
                "model": "XGBoost",
                "learning_rate": 0.01,
                "l2_reg_weight": 0.0001,
                "num_boost_rounds": 12,
            }
        )
        self.hss_1_missing_params: TParameterization = {
            "model": "Linear",
            "l2_reg_weight": 0.0001,
            "num_boost_rounds": 12,
        }
        self.hss_1_arm_missing_param = Arm(parameters=self.hss_1_missing_params)
        self.hss_1_arm_1_cast = Arm(
            parameters={
                "model": "Linear",
                "learning_rate": 0.01,
                "l2_reg_weight": 0.0001,
            }
        )
        self.hss_1_arm_2_cast = Arm(
            parameters={
                "model": "XGBoost",
                "num_boost_rounds": 12,
            }
        )

    def test_init(self) -> None:
        self.assertEqual(self.hss_1._root, self.model_parameter)
        self.assertEqual(
            self.hss_1._all_parameter_names,
            {"l2_reg_weight", "learning_rate", "num_boost_rounds", "model"},
        )
        self.assertEqual(self.hss_2._root, self.use_linear_parameter)
        self.assertEqual(
            self.hss_2._all_parameter_names,
            {
                "l2_reg_weight",
                "learning_rate",
                "num_boost_rounds",
                "model",
                "use_linear",
            },
        )

    def test_validation(self) -> None:
        # Case where dependent parameter is not in the search space.
        with self.assertRaisesRegex(ValueError, ".* 'l2_reg_weight' is not part"):
            HierarchicalSearchSpace(
                parameters=[
                    ChoiceParameter(
                        name="model",
                        parameter_type=ParameterType.STRING,
                        values=["Linear", "XGBoost"],
                        dependents={
                            "Linear": ["learning_rate", "l2_reg_weight"],
                            "XGBoost": ["num_boost_rounds"],
                        },
                    ),
                    self.lr_parameter,
                    self.num_boost_rounds_parameter,
                ]
            )

        # Case where there are two root-parameter candidates.
        with self.assertRaisesRegex(NotImplementedError, "Could not find the root"):
            HierarchicalSearchSpace(
                parameters=[
                    self.model_parameter,
                    self.model_2_parameter,
                    self.lr_parameter,
                    self.l2_reg_weight_parameter,
                    self.num_boost_rounds_parameter,
                ]
            )

        # TODO: Test case where subtrees are not independent.
        with self.assertRaisesRegex(UserInputError, ".* contain the same parameters"):
            HierarchicalSearchSpace(
                parameters=[
                    ChoiceParameter(
                        name="root",
                        parameter_type=ParameterType.BOOL,
                        values=[True, False],
                        dependents={
                            True: ["model", "model_2"],
                        },
                    ),
                    self.model_parameter,
                    self.model_2_parameter,
                    self.lr_parameter,
                    self.l2_reg_weight_parameter,
                    self.num_boost_rounds_parameter,
                ]
            )

    def test_hierarchical_structure_str(self) -> None:
        self.assertEqual(
            self.hss_1.hierarchical_structure_str(),
            f"{self.hss_1.root}\n\t(Linear)\n\t\t{self.lr_parameter}\n\t\t"
            f"{self.l2_reg_weight_parameter}\n\t(XGBoost)\n\t\t"
            f"{self.num_boost_rounds_parameter}\n",
        )
        self.assertEqual(
            self.hss_1.hierarchical_structure_str(parameter_names_only=True),
            f"{self.hss_1.root.name}\n\t(Linear)\n\t\t{self.lr_parameter.name}"
            f"\n\t\t{self.l2_reg_weight_parameter.name}\n\t(XGBoost)\n\t\t"
            f"{self.num_boost_rounds_parameter.name}\n",
        )

    def test_flatten(self) -> None:
        # Test on basic HSS.
        flattened_hss_1 = self.hss_1.flatten()
        self.assertIsNot(flattened_hss_1, self.hss_1)
        self.assertEqual(type(flattened_hss_1), SearchSpace)
        self.assertFalse(isinstance(flattened_hss_1, HierarchicalSearchSpace))
        self.assertEqual(flattened_hss_1.parameters, self.hss_1.parameters)
        self.assertEqual(
            flattened_hss_1.parameter_constraints, self.hss_1.parameter_constraints
        )
        self.assertTrue(str(self.hss_1).startswith("HierarchicalSearchSpace"))
        self.assertTrue(str(flattened_hss_1).startswith("SearchSpace"))

        # Test on HSS with constraints.
        flattened_hss_with_constraints = self.hss_with_constraints.flatten()
        self.assertIsNot(flattened_hss_with_constraints, self.hss_with_constraints)
        self.assertEqual(type(flattened_hss_with_constraints), SearchSpace)
        self.assertFalse(
            isinstance(flattened_hss_with_constraints, HierarchicalSearchSpace)
        )
        self.assertEqual(
            flattened_hss_with_constraints.parameters,
            self.hss_with_constraints.parameters,
        )
        self.assertEqual(
            flattened_hss_with_constraints.parameter_constraints,
            self.hss_with_constraints.parameter_constraints,
        )
        self.assertTrue(
            str(self.hss_with_constraints).startswith("HierarchicalSearchSpace")
        )
        self.assertTrue(str(flattened_hss_with_constraints).startswith("SearchSpace"))

    def test_cast_arm(self) -> None:
        # This uses _cast_parameterization with check_all_parameters_present=True.
        self.assertEqual(  # Check one subtree.
            self.hss_1._cast_arm(arm=self.hss_1_arm_1_flat),
            self.hss_1_arm_1_cast,
        )
        self.assertEqual(  # Check other subtree.
            self.hss_1._cast_arm(arm=self.hss_1_arm_2_flat),
            self.hss_1_arm_2_cast,
        )
        self.assertEqual(  # Check already-cast case.
            self.hss_1._cast_arm(arm=self.hss_1_arm_1_cast),
            self.hss_1_arm_1_cast,
        )
        with self.assertRaises(RuntimeError):
            self.hss_1._cast_arm(arm=self.hss_1_arm_missing_param)

    def test_cast_observation_features(self) -> None:
        # This uses _cast_parameterization with check_all_parameters_present=False.
        # Ensure that during casting, full parameterization is saved
        # in metadata and actual parameterization is cast to HSS.
        hss_1_obs_feats_1 = ObservationFeatures.from_arm(arm=self.hss_1_arm_1_flat)
        hss_1_obs_feats_1_cast = self.hss_1.cast_observation_features(
            observation_features=hss_1_obs_feats_1
        )
        self.assertEqual(  # Check one subtree.
            hss_1_obs_feats_1_cast.parameters,
            ObservationFeatures.from_arm(arm=self.hss_1_arm_1_cast).parameters,
        )
        self.assertEqual(  # Check one subtree.
            hss_1_obs_feats_1_cast.metadata.get(Keys.FULL_PARAMETERIZATION),
            hss_1_obs_feats_1.parameters,
        )
        # Check that difference with observation features made from cast arm
        # is only in metadata (to ensure only parameters and metadata are
        # manipulated during casting).
        hss_1_obs_feats_1_cast.metadata = None
        self.assertEqual(
            hss_1_obs_feats_1_cast,
            ObservationFeatures.from_arm(arm=self.hss_1_arm_1_cast),
        )

    def test_cast_parameterization(self) -> None:
        # NOTE: This is also tested in test_cast_arm & test_cast_observation_features.
        with self.assertRaisesRegex(RuntimeError, "not in parameterization to cast"):
            self.hss_1._cast_parameterization(
                parameters=self.hss_1_missing_params,
                check_all_parameters_present=True,
            )
        # An active leaf param is missing, it'll get ignored. There's an inactive
        # leaf param, that'll just get filtered out.
        self.assertEqual(
            self.hss_1._cast_parameterization(
                parameters=self.hss_1_missing_params,
                check_all_parameters_present=False,
            ),
            {"l2_reg_weight": 0.0001, "model": "Linear"},
        )
        # A hierarchical param is missing, all its dependents will be ignored.
        # In this case, it is the root param, so we'll have empty parameterization.
        self.assertEqual(
            self.hss_1._cast_parameterization(
                parameters={
                    "l2_reg_weight": 0.0001,
                    "num_boost_rounds": 12,
                },
                check_all_parameters_present=False,
            ),
            {},
        )

    def test_flatten_observation_features(self) -> None:
        # Ensure that during casting, full parameterization is saved
        # in metadata and actual parameterization is cast to HSS; during
        # flattening, parameterization in metadata is used ot inject back
        # the parameters removed during casting.
        hss_1_obs_feats_1 = ObservationFeatures.from_arm(arm=self.hss_1_arm_1_flat)
        hss_1_obs_feats_1_cast = self.hss_1.cast_observation_features(
            observation_features=hss_1_obs_feats_1
        )
        for inject_dummy in (True, False):
            hss_1_obs_feats_1_flattened = self.hss_1.flatten_observation_features(
                observation_features=hss_1_obs_feats_1_cast,
                inject_dummy_values_to_complete_flat_parameterization=inject_dummy,
            )
            self.assertEqual(  # Cast-flatten roundtrip.
                hss_1_obs_feats_1.parameters,
                hss_1_obs_feats_1_flattened.parameters,
            )
            self.assertEqual(  # Check that both cast and flattened have full params.
                hss_1_obs_feats_1_cast.metadata.get(Keys.FULL_PARAMETERIZATION),
                hss_1_obs_feats_1_flattened.metadata.get(Keys.FULL_PARAMETERIZATION),
            )
        # Check that flattening observation features without metadata does nothing.
        # Does not warn here since it already has all parameters.
        with warnings.catch_warnings(record=True) as ws:
            self.assertEqual(
                self.hss_1.flatten_observation_features(
                    observation_features=hss_1_obs_feats_1
                ),
                hss_1_obs_feats_1,
            )
        self.assertFalse(
            any("Cannot flatten observation features" in str(w.message) for w in ws)
        )
        # This one warns since it is missing some parameters.
        obs_ft_missing = ObservationFeatures.from_arm(arm=self.hss_1_arm_missing_param)
        with warnings.catch_warnings(record=True) as ws:
            self.assertEqual(
                self.hss_1.flatten_observation_features(
                    observation_features=obs_ft_missing
                ),
                obs_ft_missing,
            )
        self.assertTrue(
            any("Cannot flatten observation features" in str(w.message) for w in ws)
        )
        # Check that no warning is raised if the observation feature doesn't
        # have parameterization (for trial-index only features).
        obs_ft = ObservationFeatures(parameters={}, trial_index=0)
        with warnings.catch_warnings(record=True) as ws:
            self.assertEqual(
                self.hss_1.flatten_observation_features(observation_features=obs_ft),
                obs_ft,
            )
        self.assertFalse(
            any("Cannot flatten observation features" in str(w.message) for w in ws)
        )

    @mock.patch(f"{HierarchicalSearchSpace.__module__}.uniform", return_value=0.6)
    def test_flatten_observation_features_inject_dummy_parameter_values_with_random(
        self, mock_uniform: mock.MagicMock
    ) -> None:
        # Case 1: Linear arm
        hss_obs_feats = ObservationFeatures.from_arm(arm=self.hss_1_arm_1_cast)
        hss_obs_feats_flattened = self.hss_1.flatten_observation_features(
            observation_features=hss_obs_feats, use_random_dummy_values=True
        )
        mock_uniform.assert_not_called()
        self.assertNotIn("num_boost_rounds", hss_obs_feats_flattened.parameters)
        flattened_with_dummies = self.hss_1.flatten_observation_features(
            observation_features=hss_obs_feats,
            inject_dummy_values_to_complete_flat_parameterization=True,
            use_random_dummy_values=True,
        ).parameters
        mock_uniform.assert_called()
        self.assertIn("num_boost_rounds", flattened_with_dummies)
        self.assertEqual(
            flattened_with_dummies["num_boost_rounds"],
            1,  # int(0.6 + 0.5) = floor(0.6 + 0.5) = 1
        )
        self.assertIsInstance(  # Ensure we coerced parameter type correctly, too.
            flattened_with_dummies["num_boost_rounds"],
            int,
        )

        # Case 2: XGBoost arm
        mock_uniform.reset_mock()
        hss_obs_feats = ObservationFeatures.from_arm(arm=self.hss_1_arm_2_cast)
        hss_obs_feats_flattened = self.hss_1.flatten_observation_features(
            observation_features=hss_obs_feats, use_random_dummy_values=True
        )
        mock_uniform.assert_not_called()
        self.assertNotIn("learning_rate", hss_obs_feats_flattened.parameters)
        self.assertNotIn("l2_reg_weight", hss_obs_feats_flattened.parameters)
        flattened_with_dummies = (
            self.hss_1.flatten_observation_features(
                observation_features=hss_obs_feats,
                inject_dummy_values_to_complete_flat_parameterization=True,
                use_random_dummy_values=True,
            )
        ).parameters
        mock_uniform.assert_called()
        self.assertIn("learning_rate", flattened_with_dummies)
        self.assertIn("l2_reg_weight", flattened_with_dummies)
        self.assertEqual(
            flattened_with_dummies["learning_rate"],
            mock_uniform.return_value,
        )
        self.assertEqual(
            flattened_with_dummies["l2_reg_weight"],
            mock_uniform.return_value,
        )

        # Case 3: test setting of choice parameters
        with mock.patch(
            f"{HierarchicalSearchSpace.__module__}.choice", wraps=choice
        ) as mock_choice:
            flattened_only_dummies = self.hss_2.flatten_observation_features(
                observation_features=ObservationFeatures(
                    parameters={"num_boost_rounds": 12}
                ),
                inject_dummy_values_to_complete_flat_parameterization=True,
                use_random_dummy_values=True,
            ).parameters
            self.assertEqual(
                mock_choice.call_args_list,
                [mock.call([False, True]), mock.call(["Linear", "XGBoost"])],
            )
        self.assertEqual(
            set(flattened_only_dummies.keys()), set(self.hss_2.parameters.keys())
        )

        # Case 4: test setting of fixed parameters
        flattened_only_dummies = self.hss_with_fixed.flatten_observation_features(
            observation_features=ObservationFeatures(parameters={"use_linear": True}),
            inject_dummy_values_to_complete_flat_parameterization=True,
            use_random_dummy_values=True,
        ).parameters
        self.assertEqual(
            set(flattened_only_dummies.keys()),
            set(self.hss_with_fixed.parameters.keys()),
        )

    def test_flatten_observation_features_inject_dummy_parameter_values_non_random(
        self,
    ) -> None:
        """Check behavior with different parameter cases:
        Choice, Int-Range, Log-Range & Logit-Range.

        HSS2 has enough parameters that we can modify to test all cases.
        - `use_linear` is bool-Choice
        - `model` is string-Choice
        - `learning_rate` is Range, can be made log-scale
        - `l2_reg_weight` is Range, can be made logit-scale
        - `num_boost_rounds` is Int-Range.
        """
        checked_cast(
            RangeParameter, self.hss_2.parameters["learning_rate"]
        )._log_scale = True
        checked_cast(
            RangeParameter, self.hss_2.parameters["l2_reg_weight"]
        )._logit_scale = True
        # This has no other parameters on it, so they should all be set to
        # middle value in their respective domains.
        obs_ft = ObservationFeatures(parameters={"use_linear": False})
        flat_obs_ft = self.hss_2.flatten_observation_features(
            observation_features=obs_ft,
            inject_dummy_values_to_complete_flat_parameterization=True,
            use_random_dummy_values=False,
        )
        expected_parameters = {
            "use_linear": False,
            "model": "XGBoost",  # has two values, so it will be index 1.
            "learning_rate": 0.01,  # middle of 0.1 & 0.001 in log-scale.
            "l2_reg_weight": 0.00010004052867652256,  # middle of range in logit-scale.
            "num_boost_rounds": 15,  # mid-point of 10 & 20.
        }
        self.assertDictEqual(flat_obs_ft.parameters, expected_parameters)

    def test_flatten_observation_features_full_and_dummy(self) -> None:
        # Test flattening when both full features & inject dummy values
        # are specified. This is relevant if the full parameterization
        # is from some subset of the search space.
        # Get an obs feat with hss_1 parameterization in the metadata.
        hss_1_obs_feats_1 = ObservationFeatures.from_arm(arm=self.hss_1_arm_1_flat)
        hss_1_obs_feats_1_cast = self.hss_1.cast_observation_features(
            observation_features=hss_1_obs_feats_1
        )
        hss_1_flat_params = hss_1_obs_feats_1.parameters
        # Flatten it using hss_2, which requires an additional parameter.
        # This will work but miss a parameter.
        self.assertEqual(
            self.hss_2.flatten_observation_features(
                observation_features=hss_1_obs_feats_1_cast,
                inject_dummy_values_to_complete_flat_parameterization=False,
            ).parameters,
            hss_1_flat_params,
        )
        # Now try with inject dummy. This will add the mising param.
        hss_2_flat = self.hss_2.flatten_observation_features(
            observation_features=hss_1_obs_feats_1_cast,
            inject_dummy_values_to_complete_flat_parameterization=True,
        ).parameters
        self.assertNotEqual(hss_2_flat, hss_1_flat_params)
        self.assertEqual(
            {k: hss_2_flat[k] for k in hss_1_flat_params}, hss_1_flat_params
        )
        self.assertEqual(set(hss_2_flat.keys()), set(self.hss_2.parameters.keys()))


class TestRobustSearchSpace(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.a = RangeParameter(
            name="a", parameter_type=ParameterType.FLOAT, lower=0.5, upper=5.5
        )
        self.b = RangeParameter(
            name="b", parameter_type=ParameterType.INT, lower=2, upper=10
        )
        self.c = ChoiceParameter(
            name="c", parameter_type=ParameterType.STRING, values=["foo", "bar", "baz"]
        )
        self.parameters = [self.a, self.b, self.c]
        self.ab_dist = ParameterDistribution(
            parameters=["a", "b"],
            distribution_class="multivariate_normal",
            distribution_parameters={},
        )
        self.constraints = [
            OrderConstraint(lower_parameter=self.a, upper_parameter=self.b)
        ]
        self.rss1 = RobustSearchSpace(
            parameters=self.parameters,
            parameter_distributions=[self.ab_dist],
            parameter_constraints=cast(List[ParameterConstraint], self.constraints),
            num_samples=4,
        )

    def test_init_and_properties(self) -> None:
        # Setup some parameters and distributions.
        a_dist = ParameterDistribution(
            parameters=["a"],
            distribution_class="norm",
            distribution_parameters={"loc": 0.0, "scale": 1.0},
        )
        b_dist = ParameterDistribution(
            parameters=["b"],
            distribution_class="binom",
            distribution_parameters={"n": 2, "p": 0.3},
        )
        env1 = RangeParameter(
            name="env1", parameter_type=ParameterType.FLOAT, lower=0.5, upper=5.5
        )
        env2 = RangeParameter(
            name="env2", parameter_type=ParameterType.INT, lower=2.0, upper=10.0
        )
        choice_dist = ParameterDistribution(
            parameters=["c"],
            distribution_class="binom",
            distribution_parameters={"n": 2, "p": 0.3},
        )
        env1_dist = ParameterDistribution(
            parameters=["env1"],
            distribution_class="norm",
            distribution_parameters={"loc": 0.0, "scale": 1.0},
        )
        env1_dist_mul = ParameterDistribution(
            parameters=["env1"],
            distribution_class="norm",
            distribution_parameters={"loc": 0.0, "scale": 1.0},
            multiplicative=True,
        )
        env2_dist = ParameterDistribution(
            parameters=["env2"],
            distribution_class="binom",
            distribution_parameters={"n": 2, "p": 0.3},
        )
        mixed_dist = ParameterDistribution(
            parameters=["a", "env1"],
            distribution_class="multivariate_normal",
            distribution_parameters={"mean": [0.0, 0.0]},
        )
        # Error handling.
        with self.assertRaisesRegex(UserInputError, "Use SearchSpace instead."):
            RobustSearchSpace(
                parameters=self.parameters,
                parameter_distributions=[],
                num_samples=4,
            )
        with self.assertRaisesRegex(UserInputError, "positive integer"):
            RobustSearchSpace(
                parameters=self.parameters,
                parameter_distributions=[self.ab_dist],
                num_samples=-1,
            )
        with self.assertRaisesRegex(UnsupportedError, "all multiplicative"):
            mul_a_dist = a_dist.clone()
            mul_a_dist.multiplicative = True
            RobustSearchSpace(
                parameters=self.parameters,
                parameter_distributions=[mul_a_dist, b_dist],
                num_samples=4,
            )
        with self.assertRaisesRegex(UserInputError, "must be unique"):
            RobustSearchSpace(
                parameters=self.parameters,
                parameter_distributions=[env1_dist],
                num_samples=4,
                environmental_variables=[env1, env1],
            )
        with self.assertRaisesRegex(UserInputError, "must have a distribution"):
            RobustSearchSpace(
                parameters=self.parameters,
                parameter_distributions=[env1_dist],
                environmental_variables=[env1, env2],
                num_samples=4,
            )
        with self.assertRaisesRegex(UserInputError, "should not be repeated"):
            RobustSearchSpace(
                parameters=self.parameters,
                parameter_distributions=[a_dist],
                num_samples=4,
                environmental_variables=[self.a],
            )
        with self.assertRaisesRegex(
            UserInputError, "Distributions of environmental variables"
        ):
            RobustSearchSpace(
                parameters=self.parameters,
                parameter_distributions=[env1_dist_mul],
                num_samples=4,
                environmental_variables=[env1],
            )
        with self.assertRaisesRegex(UserInputError, "multiple parameter distributions"):
            RobustSearchSpace(
                parameters=self.parameters,
                parameter_distributions=[a_dist, a_dist],
                num_samples=4,
            )
        with self.assertRaisesRegex(UserInputError, "distribution must be"):
            RobustSearchSpace(
                parameters=self.parameters,
                parameter_distributions=[a_dist, choice_dist],
                num_samples=4,
                # pyre-fixme[6]: For 4th param expected
                #  `Optional[List[ParameterConstraint]]` but got
                #  `List[OrderConstraint]`.
                parameter_constraints=self.constraints,
            )
        with self.assertRaisesRegex(UnsupportedError, "Mixing the distribution"):
            RobustSearchSpace(
                parameters=self.parameters,
                parameter_distributions=[mixed_dist],
                num_samples=4,
                environmental_variables=[env1],
                # pyre-fixme[6]: For 5th param expected
                #  `Optional[List[ParameterConstraint]]` but got
                #  `List[OrderConstraint]`.
                parameter_constraints=self.constraints,
            )
        # Test with environmental variables.
        rss = RobustSearchSpace(
            parameters=self.parameters,
            parameter_distributions=[env1_dist, env2_dist],
            num_samples=4,
            environmental_variables=[env1, env2],
            # pyre-fixme[6]: For 5th param expected
            #  `Optional[List[ParameterConstraint]]` but got `List[OrderConstraint]`.
            parameter_constraints=self.constraints,
        )
        self.assertEqual(rss.num_samples, 4)
        self.assertTrue(rss.is_robust)
        self.assertEqual(rss.parameter_constraints, self.constraints)
        self.assertEqual(
            rss.parameters,
            {
                "a": self.a,
                "b": self.b,
                "c": self.c,
                "env1": env1,
                "env2": env2,
            },
        )
        self.assertEqual(rss.parameter_distributions, [env1_dist, env2_dist])
        self.assertEqual(rss._environmental_distributions, [env1_dist, env2_dist])
        self.assertEqual(rss._perturbation_distributions, [])
        self.assertFalse(rss.multiplicative)
        self.assertEqual(rss._distributional_parameters, {"env1", "env2"})
        self.assertEqual(rss._environmental_variables, {"env1": env1, "env2": env2})
        self.assertTrue(all(rss.is_environmental_variable(p) for p in ["env1", "env2"]))
        # Test having both types together.
        mul_a_dist = a_dist.clone()
        mul_a_dist.multiplicative = True
        rss = RobustSearchSpace(
            parameters=self.parameters,
            parameter_distributions=[mul_a_dist, env1_dist],
            num_samples=4,
            environmental_variables=[env1],
        )
        self.assertEqual(
            rss.parameters,
            {
                "a": self.a,
                "b": self.b,
                "c": self.c,
                "env1": env1,
            },
        )
        self.assertEqual(rss.parameter_distributions, [mul_a_dist, env1_dist])
        self.assertEqual(rss._environmental_distributions, [env1_dist])
        self.assertEqual(rss._perturbation_distributions, [mul_a_dist])
        self.assertTrue(rss.multiplicative)
        self.assertEqual(rss._distributional_parameters, {"a", "env1"})
        self.assertEqual(rss._environmental_variables, {"env1": env1})
        # Test with input noise.
        rss = RobustSearchSpace(
            parameters=self.parameters,
            parameter_distributions=[a_dist, b_dist],
            num_samples=4,
            # pyre-fixme[6]: For 4th param expected
            #  `Optional[List[ParameterConstraint]]` but got `List[OrderConstraint]`.
            parameter_constraints=self.constraints,
        )
        self.assertTrue(rss.is_robust)
        self.assertEqual(rss.parameter_constraints, self.constraints)
        self.assertEqual(
            rss.parameters,
            {
                "a": self.a,
                "b": self.b,
                "c": self.c,
            },
        )
        self.assertEqual(rss.parameter_distributions, [a_dist, b_dist])
        self.assertEqual(rss._environmental_distributions, [])
        self.assertEqual(rss._perturbation_distributions, [a_dist, b_dist])
        self.assertFalse(rss.multiplicative)
        self.assertEqual(rss._distributional_parameters, {"a", "b"})
        self.assertEqual(rss._environmental_variables, {})
        # Tests with a multivariate distribution.
        rss = self.rss1
        self.assertEqual(rss.parameter_constraints, self.constraints)
        self.assertEqual(
            rss.parameters,
            {
                "a": self.a,
                "b": self.b,
                "c": self.c,
            },
        )
        self.assertEqual(rss.parameter_distributions, [self.ab_dist])
        self.assertEqual(rss._environmental_distributions, [])
        self.assertEqual(rss._perturbation_distributions, [self.ab_dist])
        self.assertEqual(rss._distributional_parameters, {"a", "b"})
        self.assertEqual(rss._environmental_variables, {})
        self.assertFalse(
            any(rss.is_environmental_variable(p) for p in rss.parameters.keys())
        )

    def test_update_parameter(self) -> None:
        rss = self.rss1
        with self.assertRaisesRegex(UnsupportedError, "update_parameter"):
            rss.update_parameter(self.a)

    def test_clone(self) -> None:
        rss_clone = self.rss1.clone()
        self.assertEqual(
            rss_clone._environmental_variables, self.rss1._environmental_variables
        )
        self.assertEqual(rss_clone.parameters, self.rss1.parameters)
        self.assertEqual(rss_clone.num_samples, self.rss1.num_samples)
        self.assertEqual(
            rss_clone.parameter_constraints, self.rss1.parameter_constraints
        )
        self.assertEqual(
            rss_clone.parameter_distributions, self.rss1.parameter_distributions
        )
        self.assertEqual(
            rss_clone._distributional_parameters, self.rss1._distributional_parameters
        )

    def test_repr(self) -> None:
        expected = (
            "RobustSearchSpace("
            "parameters=["
            "RangeParameter(name='a', parameter_type=FLOAT, range=[0.5, 5.5]), "
            "RangeParameter(name='b', parameter_type=INT, range=[2, 10]), "
            "ChoiceParameter(name='c', parameter_type=STRING, "
            "values=['foo', 'bar', 'baz'], is_ordered=False, sort_values=False)], "
            "parameter_distributions=["
            "ParameterDistribution(parameters=['a', 'b'], "
            "distribution_class=multivariate_normal, distribution_parameters={}, "
            "multiplicative=False)], "
            "num_samples=4, "
            "environmental_variables=[], "
            "parameter_constraints=[OrderConstraint(a <= b)])"
        )
        self.assertEqual(str(self.rss1), expected)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from collections import defaultdict
from typing import Dict, Iterable, Set

from ax.core.base_trial import BaseTrial, TrialStatus


class SingleRunningTrialMixin:
    """Mixin for Runners with a single running trial.

    This mixin implements a simple poll_trial_status method that
    allows for a single running trial (the latest running trial).
    The returned status of trials that currently are marked as
    running is completed.
    """

    def poll_trial_status(
        self, trials: Iterable[BaseTrial]
    ) -> Dict[TrialStatus, Set[int]]:
        """Checks the status of any non-terminal trials and returns their
        indices as a mapping from TrialStatus to a list of indices. Required
        for runners used with Ax ``Scheduler``.

        NOTE: Does not need to handle waiting between polling calls while trials
        are running; this function should just perform a single poll.

        Args:
            trials: Trials to poll.

        Returns:
            A dictionary mapping TrialStatus to a list of trial indices that have
            the respective status at the time of the polling. This does not need to
            include trials that at the time of polling already have a terminal
            (ABANDONED, FAILED, COMPLETED) status (but it may).
        """
        trials = list(trials)
        if len(trials) == 0:
            return {}
        trial_statuses = defaultdict(set)
        running_trial_indices = trials[0].experiment.running_trial_indices
        max_running_trial_index = (
            -1 if len(running_trial_indices) == 0 else max(running_trial_indices)
        )
        for trial in trials:
            if trial.status in (
                TrialStatus.ABANDONED,
                TrialStatus.FAILED,
                TrialStatus.COMPLETED,
            ):
                continue
            elif (trial.status == TrialStatus.RUNNING) and (
                trial.index < max_running_trial_index
            ):
                trial_statuses[TrialStatus.COMPLETED].add(trial.index)
            else:
                trial_statuses[trial.status].add(trial.index)
        return dict(trial_statuses)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from copy import deepcopy
from unittest import mock
from unittest.mock import Mock

import torch
from ax.models.torch.botorch_defaults import (
    _get_acquisition_func,
    _get_customized_covar_module,
    _get_model,
    get_and_fit_model,
    get_warping_transform,
    NO_FEASIBLE_POINTS_MESSAGE,
)
from ax.utils.common.testutils import TestCase
from ax.utils.common.typeutils import checked_cast, not_none
from ax.utils.testing.mock import fast_botorch_optimize
from botorch.acquisition.logei import (
    qLogExpectedImprovement,
    qLogNoisyExpectedImprovement,
)
from botorch.acquisition.monte_carlo import (
    qExpectedImprovement,
    qNoisyExpectedImprovement,
    qProbabilityOfImprovement,
    qSimpleRegret,
)
from botorch.acquisition.objective import ConstrainedMCObjective
from botorch.acquisition.penalized import L1PenaltyObjective, PenalizedMCObjective
from botorch.exceptions.errors import UnsupportedError
from botorch.models.gp_regression import SingleTaskGP
from botorch.models.gp_regression_fidelity import SingleTaskMultiFidelityGP
from botorch.models.multitask import MultiTaskGP
from botorch.models.transforms.input import Warp
from botorch.utils.constraints import get_outcome_constraint_transforms
from gpytorch.kernels import MaternKernel, ScaleKernel
from gpytorch.likelihoods.gaussian_likelihood import (
    FixedNoiseGaussianLikelihood,
    GaussianLikelihood,
)
from gpytorch.module import Module
from gpytorch.priors import GammaPrior
from gpytorch.priors.lkj_prior import LKJCovariancePrior
from gpytorch.priors.prior import Prior


class BotorchDefaultsTest(TestCase):
    def test_get_model(self) -> None:
        x = torch.rand(2, 2)
        y = torch.rand(2, 1)
        var = torch.rand(2, 1)
        partial_var = torch.tensor([0, float("nan")]).unsqueeze(-1)
        unknown_var = torch.tensor([float("nan"), float("nan")]).unsqueeze(-1)
        model = _get_model(x, y, unknown_var, None)
        self.assertIsInstance(model, SingleTaskGP)
        self.assertIsInstance(model.likelihood, GaussianLikelihood)

        model = _get_model(X=x, Y=y, Yvar=var)
        self.assertIsInstance(model, SingleTaskGP)
        self.assertIsInstance(model.likelihood, FixedNoiseGaussianLikelihood)
        self.assertEqual(
            model.covar_module.base_kernel.lengthscale_prior.concentration, 3.0
        )
        self.assertEqual(model.covar_module.base_kernel.lengthscale_prior.rate, 6.0)
        model = _get_model(X=x, Y=y, Yvar=unknown_var, task_feature=1)
        self.assertIs(type(model), MultiTaskGP)  # Don't accept subclasses.
        self.assertIsInstance(model.likelihood, GaussianLikelihood)
        model = _get_model(X=x, Y=y, Yvar=var, task_feature=1)
        self.assertIsInstance(model, MultiTaskGP)
        self.assertIsInstance(model.likelihood, FixedNoiseGaussianLikelihood)
        model = _get_model(X=x, Y=y, Yvar=partial_var.clone(), task_feature=1)
        self.assertIsInstance(model, MultiTaskGP)
        model = _get_model(X=x, Y=y, Yvar=partial_var.clone(), task_feature=1, rank=1)
        self.assertEqual(model._rank, 1)
        with self.assertRaises(ValueError):
            model = _get_model(X=x, Y=y, Yvar=partial_var, task_feature=None)
        model = _get_model(X=x, Y=y, Yvar=var, fidelity_features=[-1])
        self.assertTrue(isinstance(model, SingleTaskMultiFidelityGP))
        with self.assertRaises(NotImplementedError):
            _get_model(X=x, Y=y, Yvar=var, task_feature=1, fidelity_features=[-1])
        # test fixed prior
        kwargs = {
            "prior": {
                "type": LKJCovariancePrior,
                "sd_prior": GammaPrior(2.0, 0.44),
                "eta": 0.6,
            }
        }
        # pyre-fixme[6]: For 5th param expected `Optional[List[int]]` but got
        #  `Dict[str, Union[Type[LKJCovariancePrior], float, GammaPrior]]`.
        # pyre-fixme[6]: For 5th param expected `bool` but got `Dict[str,
        #  Union[Type[LKJCovariancePrior], float, GammaPrior]]`.
        model = _get_model(X=x, Y=y, Yvar=partial_var.clone(), task_feature=1, **kwargs)
        self.assertIsInstance(
            model.task_covar_module.IndexKernelPrior,
            LKJCovariancePrior,
        )
        self.assertEqual(
            model.task_covar_module.IndexKernelPrior.sd_prior.concentration, 2.0
        )
        self.assertEqual(model.task_covar_module.IndexKernelPrior.sd_prior.rate, 0.44)
        self.assertEqual(
            model.task_covar_module.IndexKernelPrior.correlation_prior.eta, 0.6
        )

        kwargs2 = {"prior": {"type": LKJCovariancePrior}}
        model = _get_model(
            X=x,
            Y=y,
            Yvar=partial_var.clone(),
            task_feature=1,
            # pyre-fixme[6]: For 5th param expected `Optional[List[int]]` but got
            #  `Dict[str, Type[LKJCovariancePrior]]`.
            # pyre-fixme[6]: For 5th param expected `bool` but got `Dict[str,
            #  Type[LKJCovariancePrior]]`.
            **kwargs2,
        )
        self.assertIsInstance(
            model.task_covar_module.IndexKernelPrior,
            LKJCovariancePrior,
        )
        self.assertEqual(
            model.task_covar_module.IndexKernelPrior.sd_prior.concentration,
            1.0,
        )
        self.assertEqual(model.task_covar_module.IndexKernelPrior.sd_prior.rate, 0.15)
        self.assertEqual(
            model.task_covar_module.IndexKernelPrior.correlation_prior.eta,
            0.5,
        )
        kwargs3 = {
            "prior": {
                "type": LKJCovariancePrior,
                "sd_prior": GammaPrior(2.0, 0.44),
                "eta": "hi",
            }
        }
        with self.assertRaises(ValueError):
            # pyre-fixme[6]: For 5th param expected `Optional[List[int]]` but got
            #  `Dict[str, Union[Type[LKJCovariancePrior], GammaPrior, str]]`.
            # pyre-fixme[6]: For 5th param expected `bool` but got `Dict[str,
            #  Union[Type[LKJCovariancePrior], GammaPrior, str]]`.
            _get_model(X=x, Y=y, Yvar=partial_var.clone(), task_feature=1, **kwargs3)

        kwargs5 = {
            "prior": {"type": Prior, "sd_prior": GammaPrior(2.0, 0.44), "eta": 0.5}
        }
        with self.assertRaises(NotImplementedError):
            # pyre-fixme[6]: For 5th param expected `Optional[List[int]]` but got
            #  `Dict[str, Union[Type[Prior], float, GammaPrior]]`.
            # pyre-fixme[6]: For 5th param expected `bool` but got `Dict[str,
            #  Union[Type[Prior], float, GammaPrior]]`.
            _get_model(X=x, Y=y, Yvar=partial_var.clone(), task_feature=1, **kwargs5)
        # test passing customized prior
        kwargs6 = {
            "prior": {
                "covar_module_prior": {"lengthscale_prior": GammaPrior(12.0, 2.0)},
                "type": LKJCovariancePrior,
            }
        }
        model = _get_model(X=x, Y=y, Yvar=var, **deepcopy(kwargs6))  # pyre-ignore
        self.assertIsInstance(model, SingleTaskGP)
        self.assertEqual(
            model.covar_module.base_kernel.lengthscale_prior.concentration, 12.0
        )
        self.assertEqual(model.covar_module.base_kernel.lengthscale_prior.rate, 2.0)
        model = _get_model(
            X=x,
            Y=y,
            Yvar=unknown_var,
            task_feature=1,
            **deepcopy(kwargs6),  # pyre-ignore
        )
        self.assertIs(type(model), MultiTaskGP)
        self.assertIsInstance(model.likelihood, GaussianLikelihood)
        self.assertEqual(
            model.covar_module.base_kernel.lengthscale_prior.concentration, 12.0
        )
        self.assertEqual(model.covar_module.base_kernel.lengthscale_prior.rate, 2.0)
        self.assertIsInstance(
            model.task_covar_module.IndexKernelPrior,
            LKJCovariancePrior,
        )
        model = _get_model(
            X=x, Y=y, Yvar=var, task_feature=1, **deepcopy(kwargs6)  # pyre-ignore
        )
        self.assertIsInstance(model, MultiTaskGP)
        self.assertIsInstance(model.likelihood, FixedNoiseGaussianLikelihood)
        self.assertEqual(
            model.covar_module.base_kernel.lengthscale_prior.concentration, 12.0
        )
        self.assertEqual(model.covar_module.base_kernel.lengthscale_prior.rate, 2.0)
        self.assertIsInstance(
            model.task_covar_module.IndexKernelPrior, LKJCovariancePrior
        )
        # test passing customized prior
        kwargs7 = {
            "prior": {
                "covar_module_prior": {"lengthscale_prior": GammaPrior(12.0, 2.0)},
            }
        }
        covar_module = MaternKernel(
            nu=2.5,
            ard_num_dims=2,
            lengthscale_prior=GammaPrior(6.0, 6.0),
        )
        model = _get_model(
            X=x, Y=y, Yvar=var, covar_module=covar_module, **kwargs7  # pyre-ignore
        )
        self.assertIsInstance(model, SingleTaskGP)
        self.assertIsInstance(model.likelihood, FixedNoiseGaussianLikelihood)
        self.assertEqual(covar_module, model.covar_module)

        # test input warping dimension checks.
        with self.assertRaisesRegex(UnsupportedError, "batched multi output models"):
            _get_model(
                X=torch.ones(4, 3, 2),
                Y=torch.ones(4, 3, 2),
                Yvar=torch.zeros(4, 3, 2),
                use_input_warping=True,
            )

    @mock.patch("ax.models.torch.botorch_defaults._get_model", wraps=_get_model)
    @fast_botorch_optimize
    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def test_task_feature(self, get_model_mock):
        x = [torch.zeros(2, 2)]
        y = [torch.zeros(2, 1)]
        yvars = [torch.ones(2, 1)]
        get_and_fit_model(
            Xs=x,
            Ys=y,
            Yvars=yvars,
            task_features=[1],
            fidelity_features=[],
            metric_names=["L2NormMetric"],
            state_dict=None,
            refit_model=False,
        )
        # Check that task feature was correctly passed to _get_model
        self.assertEqual(get_model_mock.mock_calls[0][2]["task_feature"], 1)

        # check error on multiple task features
        with self.assertRaises(NotImplementedError):
            get_and_fit_model(
                Xs=x,
                Ys=y,
                Yvars=yvars,
                task_features=[0, 1],
                fidelity_features=[],
                metric_names=["L2NormMetric"],
                state_dict=None,
                refit_model=False,
            )

        # check error on multiple fidelity features
        with self.assertRaises(NotImplementedError):
            get_and_fit_model(
                Xs=x,
                Ys=y,
                Yvars=yvars,
                task_features=[],
                fidelity_features=[-1, -2],
                metric_names=["L2NormMetric"],
                state_dict=None,
                refit_model=False,
            )

        # check error on botch task and fidelity feature
        with self.assertRaises(NotImplementedError):
            get_and_fit_model(
                Xs=x,
                Ys=y,
                Yvars=yvars,
                task_features=[1],
                fidelity_features=[-1],
                metric_names=["L2NormMetric"],
                state_dict=None,
                refit_model=False,
            )

    @mock.patch("ax.models.torch.botorch_defaults._get_model", wraps=_get_model)
    @fast_botorch_optimize
    def test_pass_customized_prior(self, get_model_mock: Mock) -> None:
        x = [torch.zeros(2, 2)]
        y = [torch.zeros(2, 1)]
        yvars = [torch.ones(2, 1)]
        kwarg = {
            "prior": {
                "covar_module_prior": {
                    "lengthscale_prior": GammaPrior(12.0, 2.0),
                    "outputscale_prior": GammaPrior(2.0, 12.0),
                },
            }
        }
        model = get_and_fit_model(
            Xs=x,
            Ys=y,
            Yvars=yvars,
            task_features=[],
            fidelity_features=[],
            metric_names=["L2NormMetric"],
            state_dict=None,
            refit_model=False,
            **kwarg,  # pyre-ignore
        )
        self.assertIsInstance(model, SingleTaskGP)
        self.assertIsInstance(model.likelihood, FixedNoiseGaussianLikelihood)

        self.assertEqual(
            model.covar_module.base_kernel.lengthscale_prior.concentration,
            12.0,
        )
        self.assertEqual(model.covar_module.base_kernel.lengthscale_prior.rate, 2.0)
        self.assertEqual(model.covar_module.outputscale_prior.concentration, 2.0)
        self.assertEqual(model.covar_module.outputscale_prior.rate, 12.0)

        model = get_and_fit_model(
            Xs=x + x,
            Ys=y + y,
            Yvars=yvars + yvars,
            task_features=[1],
            fidelity_features=[],
            metric_names=["L2NormMetric", "L2NormMetric2"],
            state_dict=None,
            refit_model=False,
            **kwarg,  # pyre-ignore
        )
        for m in model.models:
            self.assertIs(type(m), MultiTaskGP)
            self.assertIsInstance(m.likelihood, FixedNoiseGaussianLikelihood)
            self.assertEqual(
                m.covar_module.base_kernel.lengthscale_prior.concentration,
                12.0,
            )
            self.assertEqual(m.covar_module.base_kernel.lengthscale_prior.rate, 2.0)
            self.assertEqual(m.covar_module.outputscale_prior.concentration, 2.0)
            self.assertEqual(m.covar_module.outputscale_prior.rate, 12.0)

    def test_get_acquisition_func(self) -> None:
        d, m = 3, 2
        n = 16
        x = torch.randn(n, d)
        y = torch.randn(n, m)
        unknown_var = torch.tensor([float("nan"), float("nan")]).unsqueeze(-1)
        model = _get_model(x, y, unknown_var, None)
        objective_weights = torch.tensor([1.0, 0.0])  # first output is objective
        outcome_constraints = (
            torch.tensor([[0.0, 1.0], [0.0, -1.0], [1.0, 1.0]]),  # k x m
            torch.tensor([[1.0], [-1.0], [0.0]]),  # k x 1
        )
        X_observed = torch.zeros(2, d)
        expected_constraints = not_none(
            get_outcome_constraint_transforms(outcome_constraints)
        )
        samples = torch.zeros(n, m)  # to test constraints

        for acqf_name, acqf_class in zip(
            ["qEI", "qLogEI", "qPI", "qNEI", "qLogNEI"],
            [
                qExpectedImprovement,
                qLogExpectedImprovement,
                qProbabilityOfImprovement,
                qNoisyExpectedImprovement,
                qLogNoisyExpectedImprovement,
            ],
        ):
            acqf = _get_acquisition_func(
                model=model,
                acquisition_function_name=acqf_name,
                objective_weights=objective_weights,
                outcome_constraints=outcome_constraints,
                X_observed=X_observed,
                # SampleReducingMCAcquisitionFunctions don't need this objective
                constrained_mc_objective=None,
            )
            self.assertIsInstance(acqf, acqf_class)
            acqf_constraints = acqf._constraints
            self.assertIsNotNone(acqf_constraints)

            # while the function pointer is different, return value has to be the same
            for acqf_con, exp_con in zip(acqf_constraints, expected_constraints):
                self.assertTrue(torch.allclose(acqf_con(samples), exp_con(samples)))

            with self.assertRaisesRegex(ValueError, NO_FEASIBLE_POINTS_MESSAGE):
                _get_acquisition_func(
                    model=model,
                    acquisition_function_name=acqf_name,
                    objective_weights=objective_weights,
                    outcome_constraints=outcome_constraints,
                    X_observed=None,  # errors because of no observations
                )

        # test support for PenalizedMCObjective
        penalty_objective = L1PenaltyObjective(init_point=torch.zeros(1, d))
        for acqf_name, acqf_class in zip(
            ["qEI", "qLogEI", "qNEI", "qLogNEI"],
            [
                qExpectedImprovement,
                qLogExpectedImprovement,
                qNoisyExpectedImprovement,
                qLogNoisyExpectedImprovement,
            ],
        ):
            acqf = _get_acquisition_func(
                model=model,
                acquisition_function_name=acqf_name,
                objective_weights=objective_weights,
                outcome_constraints=outcome_constraints,
                X_observed=X_observed,
                mc_objective=PenalizedMCObjective,
                constrained_mc_objective=None,
                mc_objective_kwargs={
                    "penalty_objective": penalty_objective,
                    "regularization_parameter": 0.1,
                },
            )
            self.assertIsInstance(acqf, acqf_class)
            acqf_constraints = acqf._constraints
            self.assertIsNotNone(acqf_constraints)
            self.assertIsInstance(acqf.objective, PenalizedMCObjective)
            self.assertIsInstance(acqf.objective.penalty_objective, L1PenaltyObjective)
            self.assertEqual(acqf.objective.regularization_parameter, 0.1)

        acqf_name = "qSR"
        acqf_class = qSimpleRegret
        acqf = _get_acquisition_func(
            model=model,
            acquisition_function_name=acqf_name,
            objective_weights=objective_weights,
            outcome_constraints=outcome_constraints,
            X_observed=X_observed,
            # these two need the legacy constrained objective
            constrained_mc_objective=ConstrainedMCObjective,
        )
        self.assertIsInstance(acqf, acqf_class)
        acqf_constraints = acqf._constraints
        self.assertIsNone(acqf_constraints)  # because this uses the legacy path
        self.assertIsInstance(acqf.objective, ConstrainedMCObjective)

        # the following two errors are only thrown when the acquisition function is
        # not a SampleReducingMCAcquisitionFunction.
        with self.assertRaisesRegex(
            ValueError,
            "constrained_mc_objective cannot be set to None "
            "when applying outcome constraints.",
        ):
            _get_acquisition_func(
                model=model,
                acquisition_function_name=acqf_name,
                objective_weights=objective_weights,
                outcome_constraints=outcome_constraints,
                X_observed=X_observed,
                constrained_mc_objective=None,
            )

        # these are not yet supported, will require passing additional arguments to
        # the botorch constructor (i.e. beta for UCB, ref_point and Yfor EHVI.)
        for acqf_name in ["qUCB", "qEHVI", "qNEHVI"]:
            with self.assertRaisesRegex(NotImplementedError, "not implemented yet"):
                _get_acquisition_func(
                    model=model,
                    acquisition_function_name=acqf_name,
                    objective_weights=objective_weights,
                    outcome_constraints=outcome_constraints,
                    X_observed=X_observed,
                )

    def test_get_customized_covar_module(self) -> None:
        ard_num_dims = 3
        batch_shape = torch.Size([2])
        covar_module = _get_customized_covar_module(
            covar_module_prior_dict={},
            ard_num_dims=ard_num_dims,
            aug_batch_shape=batch_shape,
            task_feature=None,
        )
        self.assertIsInstance(covar_module, Module)
        self.assertIsInstance(covar_module, ScaleKernel)
        self.assertIsInstance(covar_module.outputscale_prior, GammaPrior)
        prior = checked_cast(GammaPrior, covar_module.outputscale_prior)
        self.assertEqual(prior.concentration, 2.0)
        self.assertEqual(prior.rate, 0.15)
        self.assertIsInstance(covar_module.base_kernel, MaternKernel)
        base_kernel = checked_cast(MaternKernel, covar_module.base_kernel)
        self.assertIsInstance(base_kernel.lengthscale_prior, GammaPrior)
        self.assertEqual(
            checked_cast(GammaPrior, base_kernel.lengthscale_prior).concentration, 3.0
        )
        self.assertEqual(
            checked_cast(GammaPrior, base_kernel.lengthscale_prior).rate, 6.0
        )
        self.assertEqual(base_kernel.ard_num_dims, ard_num_dims)
        self.assertEqual(base_kernel.batch_shape, batch_shape)

        covar_module = _get_customized_covar_module(
            covar_module_prior_dict={
                "lengthscale_prior": GammaPrior(12.0, 2.0),
                "outputscale_prior": GammaPrior(2.0, 12.0),
            },
            ard_num_dims=ard_num_dims,
            aug_batch_shape=batch_shape,
            task_feature=3,
        )
        self.assertIsInstance(covar_module, Module)
        self.assertIsInstance(covar_module, ScaleKernel)
        self.assertIsInstance(covar_module.outputscale_prior, GammaPrior)
        prior = checked_cast(GammaPrior, covar_module.outputscale_prior)
        self.assertEqual(prior.concentration, 2.0)
        self.assertEqual(prior.rate, 12.0)
        self.assertIsInstance(covar_module.base_kernel, MaternKernel)
        base_kernel = checked_cast(MaternKernel, covar_module.base_kernel)
        self.assertIsInstance(base_kernel.lengthscale_prior, GammaPrior)
        self.assertEqual(
            checked_cast(GammaPrior, base_kernel.lengthscale_prior).concentration, 12.0
        )
        self.assertEqual(
            checked_cast(GammaPrior, base_kernel.lengthscale_prior).rate, 2.0
        )
        self.assertEqual(base_kernel.ard_num_dims, ard_num_dims - 1)
        self.assertEqual(base_kernel.batch_shape, batch_shape)

    def test_get_warping_transform(self) -> None:
        warp_tf = get_warping_transform(d=4)
        self.assertIsInstance(warp_tf, Warp)
        self.assertEqual(warp_tf.indices.tolist(), list(range(4)))
        warp_tf = get_warping_transform(d=4, task_feature=2)
        self.assertEqual(warp_tf.indices.tolist(), [0, 1, 3])
        warp_tf = get_warping_transform(d=4, batch_shape=torch.Size([2]))
        self.assertIsInstance(warp_tf, Warp)
        self.assertEqual(warp_tf.indices.tolist(), list(range(4)))
        self.assertEqual(warp_tf.batch_shape, torch.Size([2]))



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import ax
from ax.service.ax_client import AxClient


if __name__ == "__main__":
    assert ax is not None
    assert AxClient is not None



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import argparse
import ast
import glob
import re


def list_functions(source_glob):
    """
    List all of the functions and classes defined
    """
    defined = []
    #  Iterate through each source file
    for sp in glob.glob(source_glob):
        module_name = sp[:-3]
        module_name = module_name.replace("/", ".")
        #  Parse the source file into an AST
        node = ast.parse(open(sp).read())
        #  Extract the names of all functions and classes defined in this file
        defined.extend(
            (n.name, f"{module_name}.{n.name}")
            for n in node.body
            if (isinstance(n, ast.FunctionDef) or isinstance(n, ast.ClassDef))
        )
    return defined


def replace_backticks(source_path, docs_path):
    markdown_glob = f"{docs_path}/*.md"
    source_glob = f"{source_path}/**/*.py"
    methods = list_functions(source_glob)
    for f in glob.glob(markdown_glob):
        for n, m in methods:
            #  Match backquoted mentions of the function/class name which are
            #  not already links
            pattern = f"(?<![[`])(`{n}`)"
            link = f"[`{n}`](/api/{m.split('.')[1]}.html#{m})"
            lines = open(f).readlines()
            for i, l in enumerate(lines):
                match = re.search(pattern, l)
                if match:
                    print(f"{f}:{i+1} s/{match.group(0)}/{link}")
                    lines[i] = re.sub(pattern, link, l)
            open(f, "w").writelines(lines)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="""In markdown docs, replace backtick-quoted names of
        objects exported from Ax with links to the API docs."""
    )
    parser.add_argument(
        "--source_path",
        metavar="source_path",
        required=True,
        help="Path to source files (e.g. 'ax/').",
    )
    parser.add_argument(
        "--docs_path", type=str, required=True, help="Path to docs (e.g. 'docs/'."
    )
    args = parser.parse_args()
    replace_backticks(args.source_path, args.docs_path)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from contextlib import ExitStack
from typing import Any, Dict, Optional
from unittest import mock
from unittest.mock import Mock

import numpy as np
import torch
from ax.core.arm import Arm
from ax.core.data import Data
from ax.core.experiment import Experiment
from ax.core.metric import Metric
from ax.core.objective import Objective
from ax.core.observation import (
    Observation,
    ObservationData,
    ObservationFeatures,
    recombine_observations,
)
from ax.core.optimization_config import (
    MultiObjectiveOptimizationConfig,
    OptimizationConfig,
)
from ax.core.outcome_constraint import ScalarizedOutcomeConstraint
from ax.core.parameter import ParameterType, RangeParameter
from ax.core.search_space import SearchSpace, SearchSpaceDigest
from ax.core.types import ComparisonOp
from ax.modelbridge.base import ModelBridge
from ax.modelbridge.torch import TorchModelBridge
from ax.modelbridge.transforms.base import Transform
from ax.models.torch.botorch_modular.model import BoTorchModel
from ax.models.torch_base import TorchGenResults, TorchModel
from ax.utils.common.constants import Keys
from ax.utils.common.testutils import TestCase
from ax.utils.common.typeutils import checked_cast, not_none
from ax.utils.testing.core_stubs import (
    get_branin_data,
    get_branin_experiment,
    get_branin_search_space,
    get_experiment_with_observations,
    get_optimization_config_no_constraints,
    get_search_space_for_range_value,
)
from ax.utils.testing.mock import fast_botorch_optimize
from ax.utils.testing.modeling_stubs import get_observation1, transform_1, transform_2
from botorch.utils.datasets import (
    ContextualDataset,
    MultiTaskDataset,
    SupervisedDataset,
)


def _get_mock_modelbridge(
    dtype: Optional[torch.dtype] = None,
    device: Optional[torch.device] = None,
    fit_out_of_design: bool = False,
) -> TorchModelBridge:
    return TorchModelBridge(
        experiment=Mock(),
        search_space=Mock(),
        data=Mock(),
        model=Mock(),
        transforms=[],
        torch_dtype=dtype,
        torch_device=device,
        fit_out_of_design=fit_out_of_design,
    )


class TorchModelBridgeTest(TestCase):
    @mock.patch(
        f"{ModelBridge.__module__}.ModelBridge.__init__",
        autospec=True,
        return_value=None,
    )
    def test_TorchModelBridge(
        self,
        mock_init: Mock,
        dtype: Optional[torch.dtype] = None,
        device: Optional[torch.device] = None,
    ) -> None:
        ma = _get_mock_modelbridge(dtype=dtype, device=device)
        ma._fit_tracking_metrics = True
        ma._experiment_properties = {}
        dtype = dtype or torch.double
        self.assertEqual(ma.dtype, dtype)
        self.assertEqual(ma.device, device)
        self.assertFalse(mock_init.call_args[-1]["fit_out_of_design"])
        self.assertIsNone(ma._last_observations)
        tkwargs: Dict[str, Any] = {"dtype": dtype, "device": device}
        # Test `_fit`.
        feature_names = ["x1", "x2", "x3"]
        model = mock.MagicMock(TorchModel, autospec=True, instance=True)
        search_space = mock.MagicMock(SearchSpace, autospec=True, instance=True)
        type(search_space).parameters = mock.PropertyMock(
            return_value={fn: None for fn in feature_names}  # only need `.keys()`
        )
        X = torch.tensor([[1.0, 2.0, 3.0], [2.0, 3.0, 4.0]], **tkwargs)
        datasets = {
            "y1": SupervisedDataset(
                X=X,
                Y=torch.tensor([[3.0], [1.0]], **tkwargs),
                Yvar=torch.tensor([[4.0], [2.0]], **tkwargs),
                feature_names=feature_names,
                outcome_names=["y1"],
            ),
            "y2": SupervisedDataset(
                X=X,
                Y=torch.tensor([[2.0], [0.0]], **tkwargs),
                Yvar=torch.tensor([[2.0], [1.0]], **tkwargs),
                feature_names=feature_names,
                outcome_names=["y2"],
            ),
        }
        observation_features = [
            ObservationFeatures(parameters=dict(zip(feature_names, Xi.tolist())))
            for Xi in X
        ]
        observation_data = [
            ObservationData(
                metric_names=["y1", "y2"],
                means=np.array(y1 + y2),  # here y is already a list
                covariance=np.diag(yvar1 + yvar2),  # here yvar is already a list
            )
            for y1, y2, yvar1, yvar2 in zip(
                datasets["y1"].Y.tolist(),
                datasets["y2"].Y.tolist(),
                not_none(datasets["y1"].Yvar).tolist(),
                not_none(datasets["y2"].Yvar).tolist(),
            )
        ]
        observations = recombine_observations(observation_features, observation_data)
        ssd = SearchSpaceDigest(
            feature_names=feature_names, bounds=[(0, 1)] * 3  # pyre-ignore
        )

        with mock.patch(
            f"{TorchModelBridge.__module__}.extract_search_space_digest",
            return_value=ssd,
        ):
            ma._fit(
                model=model,
                search_space=search_space,
                observations=observations,
            )
        model_fit_args = model.fit.mock_calls[0][2]
        self.assertEqual(model_fit_args["datasets"], list(datasets.values()))
        self.assertEqual(model_fit_args["search_space_digest"], ssd)
        self.assertIsNone(model_fit_args["candidate_metadata"])
        self.assertEqual(ma._last_observations, observations)

        with mock.patch(f"{TorchModelBridge.__module__}.logger.debug") as mock_logger:
            ma._fit(
                model=model,
                search_space=search_space,
                observations=observations,
            )
        mock_logger.assert_called_once_with(
            "The observations are identical to the last set of observations "
            "used to fit the model. Skipping model fitting."
        )

        # Test `_predict`
        model.predict.return_value = (
            torch.tensor([[3.0, 2.0]], **tkwargs),
            torch.tensor([[[4.0, 0.0], [0.0, 3.0]]], **tkwargs),
        )
        pr_obs_data_expected = ObservationData(
            metric_names=["y1", "y2"],
            means=np.array([3.0, 2.0]),
            covariance=np.diag([4.0, 3.0]),
        )
        pr_obs_data = ma._predict(observation_features=observation_features[:1])
        self.assertTrue(torch.equal(model.predict.mock_calls[0][2]["X"], X[:1]))
        self.assertEqual(
            pr_obs_data,
            [pr_obs_data_expected],
        )

        # Test `_gen`

        # Hack in some properties set in the (mocked) `Modelbridge.__init__`
        ma.transforms = {}
        ma.fit_time_since_gen = 0.0
        ma._arms_by_signature = {}
        ma._fit_out_of_design = False
        ma._model_key = None
        ma._model_kwargs = None
        ma._bridge_kwargs = None

        model.gen.return_value = TorchGenResults(
            points=torch.tensor([[1.0, 2.0, 3.0]], **tkwargs),
            weights=torch.tensor([1.0], **tkwargs),
            gen_metadata={"foo": 99},
        )

        with ExitStack() as es:
            es.enter_context(
                mock.patch(
                    f"{TorchModelBridge.__module__}.extract_search_space_digest",
                    return_value=ssd,
                )
            )
            es.enter_context(
                mock.patch(
                    f"{ModelBridge.__module__}.clamp_observation_features",
                    side_effect=lambda of, ss: of,
                )
            )
            es.enter_context(
                mock.patch(
                    f"{ModelBridge.__module__}.ModelBridge._get_serialized_model_state",
                    return_value={},
                )
            )
            es.enter_context(
                mock.patch(
                    f"{TorchModelBridge.__module__}.TorchModelBridge."
                    "_array_callable_to_tensor_callable",
                    return_value=torch.round,
                )
            )
            gen_run = ma.gen(
                n=3,
                search_space=search_space,
                optimization_config=OptimizationConfig(
                    objective=Objective(metric=Metric("y1"), minimize=False),
                ),
                pending_observations={
                    "y2": [
                        ObservationFeatures(
                            parameters={"x1": 1.0, "x2": 2.0, "x3": 3.0}
                        )
                    ]
                },
                fixed_features=ObservationFeatures(parameters={"x2": 3.0}),
                model_gen_options={"option": "yes"},
            )
        gen_args = model.gen.mock_calls[0][2]
        self.assertEqual(gen_args["n"], 3)
        self.assertEqual(gen_args["search_space_digest"].bounds, [(0, 1)] * 3)
        gen_opt_config = gen_args["torch_opt_config"]
        self.assertTrue(
            torch.equal(
                gen_opt_config.objective_weights,
                torch.tensor([1.0, 0.0], **tkwargs),
            )
        )
        self.assertIsNone(gen_opt_config.outcome_constraints)
        self.assertIsNone(gen_opt_config.linear_constraints)
        self.assertEqual(gen_opt_config.fixed_features, {1: 3.0})
        X_pending_y1, X_pending_y2 = gen_opt_config.pending_observations
        self.assertTrue(torch.equal(X_pending_y1, torch.tensor([], **tkwargs)))
        self.assertTrue(
            torch.equal(X_pending_y2, torch.tensor([[1.0, 2.0, 3.0]], **tkwargs))
        )
        self.assertEqual(gen_opt_config.model_gen_options, {"option": "yes"})
        self.assertIs(gen_opt_config.rounding_func, torch.round)
        self.assertFalse(gen_opt_config.is_moo)
        self.assertEqual(gen_args["search_space_digest"].target_values, {})
        self.assertEqual(len(gen_run.arms), 1)
        self.assertEqual(gen_run.arms[0].parameters, {"x1": 1.0, "x2": 2.0, "x3": 3.0})
        self.assertEqual(gen_run.weights, [1.0])
        self.assertEqual(gen_run.fit_time, 0.0)
        self.assertEqual(gen_run.gen_metadata, {"foo": 99})

        # Test `_cross_validate`
        model.cross_validate.return_value = (
            torch.tensor([[3.0, 2.0]], **tkwargs),
            torch.tensor([[[4.0, 0.0], [0.0, 3.0]]], **tkwargs),
        )
        cv_obs_data_expected = ObservationData(
            metric_names=["y1", "y2"],
            means=np.array([3.0, 2.0]),
            covariance=np.diag([4.0, 3.0]),
        )
        cv_test_points = [
            ObservationFeatures(parameters={"x1": 1.0, "x2": 3.0, "x3": 2.0})
        ]
        X_test = torch.tensor([[1.0, 3.0, 2.0]], **tkwargs)

        with mock.patch(
            f"{TorchModelBridge.__module__}.extract_search_space_digest",
            return_value=ssd,
        ):
            cv_obs_data = ma._cross_validate(
                search_space=search_space,
                cv_training_data=observations,
                cv_test_points=cv_test_points,
            )
        model_cv_args = model.cross_validate.mock_calls[0][2]
        self.assertEqual(model_cv_args["datasets"], list(datasets.values()))
        self.assertTrue(torch.equal(model_cv_args["X_test"], X_test))
        self.assertEqual(model_cv_args["search_space_digest"], ssd)
        self.assertEqual(cv_obs_data, [cv_obs_data_expected])

        # Transform observations
        # This functionality is likely to be deprecated (T134940274)
        # so this is not a thorough test.
        ma.transform_observations(observations)

        # Transform observation features
        obsf = [ObservationFeatures(parameters={"x": 1.0, "y": 2.0})]
        ma.parameters = ["x", "y"]
        X = ma._transform_observation_features(obsf)
        self.assertTrue(torch.equal(X, torch.tensor([[1.0, 2.0]], **tkwargs)))
        # test fit out of design
        _get_mock_modelbridge(fit_out_of_design=True)
        self.assertTrue(mock_init.call_args[-1]["fit_out_of_design"])

    def test_TorchModelBridge_float(self) -> None:
        self.test_TorchModelBridge(dtype=torch.float)

    def test_TorchModelBridge_cuda(self) -> None:
        if torch.cuda.is_available():
            self.test_TorchModelBridge(device=torch.device("cuda"))

    @mock.patch(f"{TorchModel.__module__}.TorchModel", autospec=True)
    @mock.patch(f"{ModelBridge.__module__}.ModelBridge.__init__")
    def test_evaluate_acquisition_function(self, _, mock_torch_model: Mock) -> None:
        ma = TorchModelBridge(
            experiment=Mock(),
            search_space=Mock(),
            data=Mock(),
            model=Mock(),
            transforms=[],
            torch_dtype=torch.float64,
            torch_device=torch.device("cpu"),
        )
        ma._fit_tracking_metrics = True
        ma._experiment_properties = {}
        # These attributes would've been set by `ModelBridge` __init__, but it's mocked.
        ma.model = mock_torch_model()
        t = mock.MagicMock(Transform, autospec=True, wraps=Transform(None, None, None))
        ma.transforms = {"ExampleTransform": t}
        ma.parameters = ["x", "y"]
        model_eval_acqf = mock_torch_model.return_value.evaluate_acquisition_function
        model_eval_acqf.return_value = torch.tensor([5.0], dtype=torch.float64)

        ma._model_space = get_branin_search_space()
        ma._optimization_config = None
        ma.outcomes = ["test_metric"]
        ma._fit_out_of_design = False

        with self.assertRaisesRegex(ValueError, "optimization_config"):
            ma.evaluate_acquisition_function(
                observation_features=[
                    ObservationFeatures(parameters={"x": 1.0, "y": 2.0})
                ],
            )

        with mock.patch(
            "ax.modelbridge.torch.extract_search_space_digest",
            return_value=SearchSpaceDigest(feature_names=[], bounds=[]),
        ):
            acqf_vals = ma.evaluate_acquisition_function(
                observation_features=[
                    ObservationFeatures(parameters={"x": 1.0, "y": 2.0})
                ],
                optimization_config=get_optimization_config_no_constraints(),
            )

        self.assertEqual(acqf_vals, [5.0])
        t.transform_observation_features.assert_any_call(
            [ObservationFeatures(parameters={"x": 1.0, "y": 2.0})],
        )
        t.transform_observation_features.reset_mock()
        model_eval_acqf.assert_called_once()
        self.assertTrue(
            torch.equal(  # `call_args` is an (args, kwargs) tuple
                model_eval_acqf.call_args[1]["X"],
                torch.tensor([[[1.0, 2.0]]], dtype=torch.float64),
            )
        )

        # Test evaluating at multiple points.
        # Case 1: List[ObsFeat, ObsFeat], should be 2 x 1 x d.
        with mock.patch(
            "ax.modelbridge.torch.extract_search_space_digest",
            return_value=SearchSpaceDigest(feature_names=[], bounds=[]),
        ):
            acqf_vals = ma.evaluate_acquisition_function(
                observation_features=[
                    ObservationFeatures(parameters={"x": 1.0, "y": 2.0}),
                    ObservationFeatures(parameters={"x": 1.0, "y": 2.0}),
                ],
                optimization_config=get_optimization_config_no_constraints(),
            )
        t.transform_observation_features.assert_any_call(
            [ObservationFeatures(parameters={"x": 1.0, "y": 2.0})],
        )
        t.transform_observation_features.reset_mock()
        self.assertTrue(
            torch.equal(  # `call_args` is an (args, kwargs) tuple
                model_eval_acqf.call_args[-1]["X"],
                torch.tensor([[[1.0, 2.0]], [[1.0, 2.0]]], dtype=torch.float64),
            )
        )
        # Case 2: List[List[ObsFeat, ObsFeat]], should be 1 x 2 x d.
        with mock.patch(
            "ax.modelbridge.torch.extract_search_space_digest",
            return_value=SearchSpaceDigest(feature_names=[], bounds=[]),
        ):
            acqf_vals = ma.evaluate_acquisition_function(
                observation_features=[
                    [
                        ObservationFeatures(parameters={"x": 1.0, "y": 2.0}),
                        ObservationFeatures(parameters={"x": 1.0, "y": 2.0}),
                    ]
                ],
                optimization_config=get_optimization_config_no_constraints(),
            )
        t.transform_observation_features.assert_any_call(
            [
                ObservationFeatures(parameters={"x": 1.0, "y": 2.0}),
                ObservationFeatures(parameters={"x": 1.0, "y": 2.0}),
            ],
        )
        self.assertTrue(
            torch.equal(  # `call_args` is an (args, kwargs) tuple
                model_eval_acqf.call_args[-1]["X"],
                torch.tensor([[[1.0, 2.0], [1.0, 2.0]]], dtype=torch.float64),
            )
        )

    @mock.patch(
        f"{ModelBridge.__module__}.observations_from_data",
        autospec=True,
        return_value=([get_observation1()]),
    )
    @mock.patch(
        f"{ModelBridge.__module__}.unwrap_observation_data",
        autospec=True,
        return_value=(2, 2),
    )
    @mock.patch(
        f"{ModelBridge.__module__}.gen_arms",
        autospec=True,
        return_value=([Arm(parameters={})], {}),
    )
    @mock.patch(
        f"{ModelBridge.__module__}.ModelBridge.predict",
        autospec=True,
        return_value=({"m": [1.0]}, {"m": {"m": [2.0]}}),
    )
    @mock.patch(f"{TorchModelBridge.__module__}.TorchModelBridge._fit", autospec=True)
    @mock.patch(
        f"{TorchModel.__module__}.TorchModel.gen",
        return_value=TorchGenResults(
            points=torch.tensor([[1, 2]]),
            weights=torch.tensor([1.0]),
        ),
        autospec=True,
    )
    def test_best_point(
        self,
        _mock_gen,
        _mock_fit,
        _mock_predict,
        _mock_gen_arms,
        _mock_unwrap,
        _mock_obs_from_data,
    ) -> None:
        exp = Experiment(search_space=get_search_space_for_range_value(), name="test")
        oc = OptimizationConfig(
            objective=Objective(metric=Metric("a"), minimize=False),
            outcome_constraints=[],
        )
        search_space = get_search_space_for_range_value()
        modelbridge = TorchModelBridge(
            search_space=search_space,
            model=TorchModel(),
            transforms=[transform_1, transform_2],
            experiment=exp,
            data=Data(),
            optimization_config=oc,
        )

        self.assertEqual(
            list(modelbridge.transforms.keys()),
            ["Cast", "transform_1", "transform_2"],
        )

        # _fit is mocked, which sets these
        modelbridge.parameters = list(search_space.parameters.keys())
        modelbridge.outcomes = ["a"]

        with mock.patch(
            f"{TorchModel.__module__}.TorchModel.best_point",
            return_value=torch.tensor([1.0, 2.0]),
            autospec=True,
        ):
            run = modelbridge.gen(n=1, optimization_config=oc)
            arm, predictions = not_none(run.best_arm_predictions)
            model_arm, model_predictions = not_none(modelbridge.model_best_point())
            predictions = not_none(predictions)
            model_predictions = not_none(model_predictions)
        self.assertEqual(arm.parameters, {})
        self.assertEqual(predictions[0], {"m": 1.0})
        self.assertEqual(predictions[1], {"m": {"m": 2.0}})
        self.assertEqual(model_predictions[0], {"m": 1.0})
        self.assertEqual(model_predictions[1], {"m": {"m": 2.0}})

        # test optimization config validation - raise error when
        # ScalarizedOutcomeConstraint contains a metric that is not in the outcomes
        with self.assertRaises(ValueError):
            modelbridge.gen(
                n=1,
                optimization_config=OptimizationConfig(
                    objective=Objective(metric=Metric("a"), minimize=False),
                    outcome_constraints=[
                        ScalarizedOutcomeConstraint(
                            metrics=[Metric("wrong_metric_name")],
                            weights=[1.0],
                            op=ComparisonOp.LEQ,
                            bound=0,
                        )
                    ],
                ),
            )

        with mock.patch(
            f"{TorchModel.__module__}.TorchModel.best_point",
            side_effect=NotImplementedError,
            autospec=True,
        ):
            res = modelbridge.model_best_point()
        self.assertIsNone(res)

    @mock.patch(
        f"{ModelBridge.__module__}.observations_from_data",
        autospec=True,
        return_value=([get_observation1()]),
    )
    @mock.patch(
        f"{ModelBridge.__module__}.unwrap_observation_data",
        autospec=True,
        return_value=(2, 2),
    )
    @mock.patch(
        f"{ModelBridge.__module__}.gen_arms",
        autospec=True,
        return_value=[Arm(parameters={})],
    )
    @mock.patch(
        f"{ModelBridge.__module__}.ModelBridge.predict",
        autospec=True,
        return_value=({"m": [1.0]}, {"m": {"m": [2.0]}}),
    )
    @mock.patch(f"{TorchModelBridge.__module__}.TorchModelBridge._fit", autospec=True)
    # pyre-fixme[56]: Pyre was not able to infer the type of argument
    #  `numpy.array([[[1.000000]], [[2.000000]]])` to decorator factory
    #  `unittest.mock.patch`.
    @mock.patch(
        f"{TorchModel.__module__}.TorchModel.feature_importances",
        return_value=np.array([[[1.0]], [[2.0]]]),
        autospec=True,
    )
    def test_importances(
        self,
        _mock_feature_importances,
        _mock_fit,
        _mock_predict,
        _mock_gen_arms,
        _mock_unwrap,
        _mock_obs_from_data,
    ) -> None:
        exp = Experiment(search_space=get_search_space_for_range_value(), name="test")
        search_space = get_search_space_for_range_value()
        modelbridge = TorchModelBridge(
            search_space=search_space,
            model=TorchModel(),
            transforms=[transform_1, transform_2],
            experiment=exp,
            data=Data(),
        )
        # _fit is mocked, which typically sets these
        modelbridge.parameters = list(search_space.parameters.keys())
        modelbridge.outcomes = ["a", "b"]

        self.assertEqual(modelbridge.feature_importances("a"), {"x": [1.0]})
        self.assertEqual(modelbridge.feature_importances("b"), {"x": [2.0]})

    @mock.patch(
        f"{TorchModel.__module__}.TorchModel.gen",
        return_value=TorchGenResults(
            points=torch.tensor([[1, 2], [2, 3]]),
            weights=torch.tensor([1.0, 2.0]),
            candidate_metadata=[
                {"some_key": "some_value_0"},
                {"some_key": "some_value_1"},
            ],
        ),
        autospec=True,
    )
    @mock.patch(f"{TorchModel.__module__}.TorchModel.update", autospec=True)
    @mock.patch(f"{TorchModel.__module__}.TorchModel.fit", autospec=True)
    def test_candidate_metadata_propagation(
        self,
        mock_model_fit: Mock,
        mock_model_update: Mock,
        mock_model_gen: Mock,
    ) -> None:
        exp = get_branin_experiment(with_status_quo=True, with_batch=True)
        # Check that the metadata is correctly re-added to observation
        # features during `fit`.
        # pyre-fixme[16]: `BaseTrial` has no attribute `_generator_run_structs`.
        preexisting_batch_gr = exp.trials[0]._generator_run_structs[0].generator_run
        preexisting_batch_gr._candidate_metadata_by_arm_signature = {
            preexisting_batch_gr.arms[0].signature: {
                "preexisting_batch_cand_metadata": "some_value"
            }
        }
        modelbridge = TorchModelBridge(
            experiment=exp,
            search_space=exp.search_space,
            model=TorchModel(),
            transforms=[],
            data=get_branin_data(),
        )

        datasets = mock_model_fit.call_args[1].get("datasets")
        X_expected = torch.tensor(
            [list(exp.trials[0].arms[0].parameters.values())],
            dtype=torch.double,
        )
        for dataset in datasets:
            self.assertTrue(torch.equal(dataset.X, X_expected))

        self.assertEqual(
            mock_model_fit.call_args[1].get("candidate_metadata"),
            [[{"preexisting_batch_cand_metadata": "some_value"}]],
        )

        # Check that `gen` correctly propagates the metadata to the GR.
        gr = modelbridge.gen(n=1)
        self.assertEqual(
            gr.candidate_metadata_by_arm_signature,
            {
                gr.arms[0].signature: {"some_key": "some_value_0"},
                gr.arms[1].signature: {"some_key": "some_value_1"},
            },
        )

        # Check that `None` candidate metadata is handled correctly.
        mock_model_gen.return_value = TorchGenResults(
            points=torch.tensor([[2, 4], [3, 5]]),
            weights=torch.tensor([1.0, 2.0]),
            candidate_metadata=None,
        )
        gr = modelbridge.gen(n=1)
        self.assertIsNone(gr.candidate_metadata_by_arm_signature)

        # Check that no candidate metadata is handled correctly.
        exp = get_branin_experiment(with_status_quo=True)

        with mock.patch(
            f"{TorchModelBridge.__module__}."
            "TorchModelBridge._validate_observation_data",
            autospec=True,
        ):
            modelbridge = TorchModelBridge(
                search_space=exp.search_space,
                experiment=exp,
                model=TorchModel(),
                data=Data(),
                transforms=[],
            )
            # Hack in outcome names to bypass validation (since we did not pass any
            # to the model so _fit did not populate this)
            metric_name = next(iter(exp.metrics))
            modelbridge.outcomes = [metric_name]
            modelbridge._metric_names = {metric_name}
        gr = modelbridge.gen(n=1)
        self.assertIsNone(mock_model_fit.call_args[1].get("candidate_metadata"))
        self.assertIsNone(gr.candidate_metadata_by_arm_signature)

    @mock.patch(f"{TorchModel.__module__}.TorchModel.fit", autospec=True)
    def test_fit_tracking_metrics(self, mock_model_fit: Mock) -> None:
        exp = get_experiment_with_observations(
            observations=[[0.0, 0.0, 0.0], [1.0, 1.0, 1.0], [2.0, 2.0, 2.0]],
            with_tracking_metrics=True,
        )
        for fit_tracking_metrics in (True, False):
            mock_model_fit.reset_mock()
            modelbridge = TorchModelBridge(
                experiment=exp,
                search_space=exp.search_space,
                data=exp.lookup_data(),
                model=TorchModel(),
                transforms=[],
                fit_tracking_metrics=fit_tracking_metrics,
            )
            mock_model_fit.assert_called_once()
            call_kwargs = mock_model_fit.call_args.kwargs
            if fit_tracking_metrics:
                expected_outcomes = ["m1", "m2", "m3"]
            else:
                expected_outcomes = ["m1", "m2"]
            self.assertEqual(modelbridge.outcomes, expected_outcomes)
            self.assertEqual(len(call_kwargs["datasets"]), len(expected_outcomes))

    def test_convert_observations(self) -> None:
        with mock.patch(
            f"{ModelBridge.__module__}.ModelBridge.__init__",
            autospec=True,
        ):
            mb = _get_mock_modelbridge()
            mb._experiment_properties = {"parameter_decomposition": None}
        raw_X = torch.rand(10, 3) * 5
        raw_X[:, -1].round_()  # Make sure last column is integer.
        raw_X[0, -1] = 0  # Make sure task value 0 exists.
        raw_Y = torch.sin(raw_X).sum(-1)
        feature_names = ["x0", "x1", "x2"]
        metric_names = ["y"]
        observation_features = [
            ObservationFeatures(
                parameters={feature_names[i]: x_[i].item() for i in range(3)}
            )
            for x_ in raw_X
        ]
        observation_data = [
            ObservationData(
                metric_names=metric_names,
                means=np.asarray([y]),
                covariance=np.array([[float("nan")]]),
            )
            for y in raw_Y
        ]
        for use_task, expected_class in (
            (True, MultiTaskDataset),
            (False, SupervisedDataset),
        ):
            search_space_digest = SearchSpaceDigest(
                feature_names=feature_names,
                bounds=[(0.0, 5.0)] * 3,
                ordinal_features=[2],
                discrete_choices={2: list(range(0, 11))},  # pyre-ignore
                task_features=[2] if use_task else [],
                target_values={2: 0} if use_task else {},  # pyre-ignore
            )
            converted_datasets, ordered_outcomes, _ = mb._convert_observations(
                observation_data=observation_data,
                observation_features=observation_features,
                outcomes=metric_names,
                parameters=feature_names,
                search_space_digest=search_space_digest,
            )
            self.assertEqual(len(converted_datasets), 1)
            dataset = not_none(converted_datasets[0])
            self.assertIs(dataset.__class__, expected_class)
            if use_task:
                sort_idx = torch.argsort(raw_X[:, -1])
                expected_X = raw_X[sort_idx]
                expected_Y = raw_Y[sort_idx].unsqueeze(-1)
            else:
                expected_X = raw_X
                expected_Y = raw_Y.unsqueeze(-1)
            self.assertTrue(torch.equal(dataset.X, expected_X))
            self.assertTrue(torch.equal(dataset.Y, expected_Y))
            self.assertIsNone(dataset.Yvar)
            self.assertEqual(dataset.feature_names, feature_names)
            self.assertEqual(dataset.outcome_names, metric_names)
            self.assertEqual(ordered_outcomes, metric_names)

            with self.assertRaisesRegex(ValueError, "was not observed."):
                mb._convert_observations(
                    observation_data=observation_data,
                    observation_features=observation_features,
                    outcomes=metric_names + ["extra"],
                    parameters=feature_names,
                    search_space_digest=search_space_digest,
                )

    def test_convert_contextual_observations(self) -> None:
        raw_X = torch.rand(10, 3) * 5
        raw_X[:, -1].round_()  # Make sure last column is integer.
        raw_X[0, -1] = 0  # Make sure task value 0 exists.
        raw_Y = torch.sin(raw_X).sum(-1)
        feature_names = ["x0", "x1", "x2"]
        metric_names = ["y", "y:c0", "y:c1", "y:c2"]
        parameter_decomposition = {f"c{i}": [f"x{i}"] for i in range(3)}
        metric_decomposition = {f"c{i}": [f"y:c{i}"] for i in range(3)}

        with mock.patch(
            f"{ModelBridge.__module__}.ModelBridge.__init__",
            autospec=True,
        ):
            mb = _get_mock_modelbridge()
            mb._experiment_properties = {
                "parameter_decomposition": parameter_decomposition,
                "metric_decomposition": metric_decomposition,
            }

        observation_features = [
            ObservationFeatures(
                parameters={feature_names[i]: x_[i].item() for i in range(3)}
            )
            for x_ in raw_X
        ]
        num_m = len(metric_names)
        observation_data = [
            ObservationData(
                metric_names=metric_names,
                means=np.asarray([y for _ in range(num_m)]),
                covariance=np.array(
                    [float("nan") for _ in range(num_m * num_m)]
                ).reshape([num_m, num_m]),
            )
            for y in raw_Y
        ]
        converted_datasets, ordered_outcomes, _ = mb._convert_observations(
            observation_data=observation_data,
            observation_features=observation_features,
            outcomes=metric_names,
            parameters=feature_names,
            search_space_digest=SearchSpaceDigest(
                feature_names=feature_names,
                bounds=[(0.0, 5.0)] * 3,
                ordinal_features=[2],
                discrete_choices={2: list(range(0, 11))},  # pyre-ignore
                task_features=[],
                target_values={},
            ),
        )
        self.assertEqual(len(converted_datasets), 2)
        expected_outcomes = list(converted_datasets[0].outcome_names)
        expected_outcomes.extend(list(converted_datasets[1].outcome_names))
        self.assertEqual(ordered_outcomes, expected_outcomes)
        for dataset in converted_datasets:
            self.assertIsInstance(dataset, ContextualDataset)
            self.assertEqual(dataset.feature_names, feature_names)
            self.assertDictEqual(
                checked_cast(ContextualDataset, dataset).parameter_decomposition,
                parameter_decomposition,
            )
            if len(dataset.outcome_names) == 1:
                self.assertListEqual(dataset.outcome_names, ["y"])
                self.assertTrue(torch.equal(dataset.X, raw_X))
                self.assertTrue(torch.equal(dataset.Y, raw_Y.unsqueeze(-1)))
            else:
                self.assertListEqual(dataset.outcome_names, ["y:c0", "y:c1", "y:c2"])
                self.assertListEqual(
                    checked_cast(ContextualDataset, dataset).context_buckets,
                    ["c0", "c1", "c2"],
                )
                self.assertDictEqual(
                    not_none(
                        checked_cast(ContextualDataset, dataset).metric_decomposition
                    ),
                    metric_decomposition,
                )
                self.assertTrue(torch.equal(dataset.X, raw_X))
                self.assertTrue(
                    torch.equal(
                        dataset.Y,
                        torch.cat([raw_Y.unsqueeze(-1) for _ in range(3)], dim=-1),
                    )
                )
        # Test _get_fit_args handling of outcome names
        mb._fit_tracking_metrics = True
        search_space = SearchSpace(
            parameters=[
                RangeParameter(
                    name=f"x{i}",
                    lower=0.0,
                    upper=5.0,
                    parameter_type=ParameterType.FLOAT,
                )
                for i in range(3)
            ]
        )
        observations = []
        for i, od in enumerate(observation_data):
            observations.append(Observation(data=od, features=observation_features[i]))
        converted_datasets2, _, _ = mb._get_fit_args(
            search_space=search_space,
            observations=observations,
            parameters=feature_names,
            update_outcomes_and_parameters=True,
        )
        self.assertEqual(mb.outcomes, expected_outcomes)
        self.assertEqual(converted_datasets, converted_datasets2)
        datasets, _, _ = mb._get_fit_args(
            search_space=search_space,
            observations=observations,
            parameters=feature_names,
            update_outcomes_and_parameters=False,
        )
        self.assertEqual(mb.outcomes, expected_outcomes)

    @fast_botorch_optimize
    def test_gen_metadata_untransform(self) -> None:
        experiment = get_experiment_with_observations(
            observations=[[0.0, 1.0], [2.0, 3.0]]
        )
        model = BoTorchModel()
        mb = TorchModelBridge(
            experiment=experiment,
            search_space=experiment.search_space,
            data=experiment.lookup_data(),
            model=model,
            transforms=[],
        )
        for additional_metadata in (
            {},
            {"objective_thresholds": None},
            {
                "objective_thresholds": checked_cast(
                    MultiObjectiveOptimizationConfig, experiment.optimization_config
                ).objective_thresholds,
            },
        ):
            gen_return_value = TorchGenResults(
                points=torch.tensor([[1.0, 2.0, 3.0]]),
                weights=torch.tensor([1.0]),
                gen_metadata={Keys.EXPECTED_ACQF_VAL: [1.0], **additional_metadata},
            )
            with mock.patch.object(
                mb, "_untransform_objective_thresholds"
            ) as mock_untransform, mock.patch.object(
                model, "gen", return_value=gen_return_value
            ):
                mb.gen(n=1)
            if additional_metadata.get("objective_thresholds", None) is None:
                mock_untransform.assert_not_called()
            else:
                mock_untransform.assert_called_once()



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from typing import Any, Dict

from ax.core.metric import Metric
from ax.core.optimization_config import MultiObjectiveOptimizationConfig
from ax.core.parameter import (
    ChoiceParameter,
    FixedParameter,
    ParameterType,
    RangeParameter,
)
from ax.core.search_space import HierarchicalSearchSpace
from ax.service.utils.instantiation import InstantiationBase
from ax.utils.common.testutils import TestCase
from ax.utils.common.typeutils import checked_cast


class TestInstantiationtUtils(TestCase):
    """Testing the instantiation utilities functionality that is not tested in
    main `AxClient` testing suite (`TestServiceAPI`)."""

    def test_parameter_type_validation(self) -> None:
        with self.assertRaisesRegex(ValueError, "No AE parameter type"):
            # pyre-fixme[6]: For 1st param expected `Union[Type[bool], Type[float],
            #  Type[int], Type[str]]` but got `Type[list]`.
            InstantiationBase._get_parameter_type(list)

    def test_make_search_space(self) -> None:
        with self.assertRaisesRegex(ValueError, "cannot contain spaces"):
            InstantiationBase.make_search_space(
                parameters=[
                    {
                        "name": "x space 1",
                        "type": "range",
                        "bounds": [0.0, 1.0],
                    }
                ],
                parameter_constraints=None,
            )

    def test_constraint_from_str(self) -> None:
        with self.assertRaisesRegex(ValueError, "Bound for the constraint"):
            InstantiationBase.constraint_from_str(
                "x1 + x2 <= not_numerical_bound",
                # pyre-fixme[6]: For 2nd param expected `Dict[str, Parameter]` but
                #  got `Dict[str, None]`.
                {"x1": None, "x2": None},
            )
        with self.assertRaisesRegex(ValueError, "Outcome constraint bound"):
            InstantiationBase.outcome_constraint_from_str("m1 <= not_numerical_bound")
        three_val_constaint = InstantiationBase.constraint_from_str(
            "x1 + x2 + x3 <= 3",
            {
                "x1": RangeParameter(
                    name="x1", parameter_type=ParameterType.FLOAT, lower=0.1, upper=2.0
                ),
                "x2": RangeParameter(
                    name="x2", parameter_type=ParameterType.FLOAT, lower=0.1, upper=2.0
                ),
                "x3": RangeParameter(
                    name="x3", parameter_type=ParameterType.FLOAT, lower=0.1, upper=2.0
                ),
            },
        )

        self.assertEqual(three_val_constaint.bound, 3.0)
        with self.assertRaisesRegex(ValueError, "Parameter constraint should"):
            InstantiationBase.constraint_from_str(
                "x1 + x2 + <= 3",
                # pyre-fixme[6]: For 2nd param expected `Dict[str, Parameter]` but
                #  got `Dict[str, None]`.
                {"x1": None, "x2": None, "x3": None},
            )
        with self.assertRaisesRegex(ValueError, "Parameter constraint should"):
            InstantiationBase.constraint_from_str(
                "x1 + x2 + x3 = 3",
                # pyre-fixme[6]: For 2nd param expected `Dict[str, Parameter]` but
                #  got `Dict[str, None]`.
                {"x1": None, "x2": None, "x3": None},
            )
        one_val_constraint = InstantiationBase.constraint_from_str(
            "x1 <= 0",
            # pyre-fixme[6]: For 2nd param expected `Dict[str, Parameter]` but
            #  got `Dict[str, None]`.
            {"x1": None, "x2": None},
        )
        self.assertEqual(one_val_constraint.bound, 0.0)
        self.assertEqual(one_val_constraint.constraint_dict, {"x1": 1.0})
        one_val_constraint = InstantiationBase.constraint_from_str(
            "-0.5*x1 >= -0.1",
            # pyre-fixme[6]: For 2nd param expected `Dict[str, Parameter]` but
            #  got `Dict[str, None]`.
            {"x1": None, "x2": None},
        )
        self.assertEqual(one_val_constraint.bound, 0.1)
        self.assertEqual(one_val_constraint.constraint_dict, {"x1": 0.5})
        three_val_constaint2 = InstantiationBase.constraint_from_str(
            "-x1 + 2.1*x2 - 4*x3 <= 3",
            {
                "x1": RangeParameter(
                    name="x1", parameter_type=ParameterType.FLOAT, lower=0.1, upper=4.0
                ),
                "x2": RangeParameter(
                    name="x2", parameter_type=ParameterType.FLOAT, lower=0.1, upper=4.0
                ),
                "x3": RangeParameter(
                    name="x3", parameter_type=ParameterType.FLOAT, lower=0.1, upper=4.0
                ),
            },
        )

        self.assertEqual(three_val_constaint2.bound, 3.0)
        self.assertEqual(
            three_val_constaint2.constraint_dict, {"x1": -1.0, "x2": 2.1, "x3": -4.0}
        )
        with self.assertRaisesRegex(ValueError, "Multiplier should be float"):
            InstantiationBase.constraint_from_str(
                "x1 - e*x2 + x3 <= 3",
                # pyre-fixme[6]: For 2nd param expected `Dict[str, Parameter]` but
                #  got `Dict[str, None]`.
                {"x1": None, "x2": None, "x3": None},
            )
        with self.assertRaisesRegex(ValueError, "A linear constraint should be"):
            InstantiationBase.constraint_from_str(
                "x1 - 2 *x2 + 3 *x3 <= 3",
                # pyre-fixme[6]: For 2nd param expected `Dict[str, Parameter]` but
                #  got `Dict[str, None]`.
                {"x1": None, "x2": None, "x3": None},
            )
        with self.assertRaisesRegex(ValueError, "A linear constraint should be"):
            InstantiationBase.constraint_from_str(
                "x1 - 2* x2 + 3* x3 <= 3",
                # pyre-fixme[6]: For 2nd param expected `Dict[str, Parameter]` but
                #  got `Dict[str, None]`.
                {"x1": None, "x2": None, "x3": None},
            )
        with self.assertRaisesRegex(ValueError, "A linear constraint should be"):
            InstantiationBase.constraint_from_str(
                "x1 - 2 * x2 + 3*x3 <= 3",
                # pyre-fixme[6]: For 2nd param expected `Dict[str, Parameter]` but
                #  got `Dict[str, None]`.
                {"x1": None, "x2": None, "x3": None},
            )

    def test_add_tracking_metrics(self) -> None:
        experiment = InstantiationBase.make_experiment(
            parameters=[{"name": "x", "type": "range", "bounds": [0, 1]}],
            tracking_metric_names=None,
        )
        self.assertDictEqual(experiment._tracking_metrics, {})

        metrics_names = ["metric_1", "metric_2"]
        experiment = InstantiationBase.make_experiment(
            parameters=[{"name": "x", "type": "range", "bounds": [0, 1]}],
            tracking_metric_names=metrics_names,
        )
        self.assertDictEqual(
            experiment._tracking_metrics,
            {metric_name: Metric(name=metric_name) for metric_name in metrics_names},
        )

    def test_make_objectives(self) -> None:
        with self.assertRaisesRegex(ValueError, "specify 'minimize' or 'maximize'"):
            InstantiationBase.make_objectives({"branin": "unknown"})

        with self.assertRaisesRegex(ValueError, "cannot contain spaces"):
            InstantiationBase.make_objectives({"branin space": "maximize"})

        objectives = InstantiationBase.make_objectives(
            {"branin": "minimize", "currin": "maximize"}
        )
        branin_metric = [o.minimize for o in objectives if o.metric.name == "branin"]
        self.assertTrue(branin_metric[0])
        currin_metric = [o.minimize for o in objectives if o.metric.name == "currin"]
        self.assertFalse(currin_metric[0])

    def test_make_optimization_config(self) -> None:
        objectives = {"branin": "minimize", "currin": "maximize"}
        objective_thresholds = ["branin <= 0", "currin >= 0"]
        with self.subTest("Single-objective optimizations with objective thresholds"):
            with self.assertRaisesRegex(ValueError, "not specify objective thresholds"):
                InstantiationBase.make_optimization_config(
                    {"branin": "minimize"},
                    objective_thresholds,
                    outcome_constraints=[],
                    status_quo_defined=False,
                )

        with self.subTest("MOO with partial objective thresholds"):
            multi_optimization_config = InstantiationBase.make_optimization_config(
                objectives,
                objective_thresholds=objective_thresholds[:1],
                outcome_constraints=[],
                status_quo_defined=False,
            )
            self.assertEqual(len(multi_optimization_config.objective.metrics), 2)
            self.assertEqual(
                len(
                    checked_cast(
                        MultiObjectiveOptimizationConfig, multi_optimization_config
                    ).objective_thresholds
                ),
                1,
            )

        with self.subTest("MOO with all objective threshold"):
            multi_optimization_config = InstantiationBase.make_optimization_config(
                objectives,
                objective_thresholds,
                outcome_constraints=[],
                status_quo_defined=False,
            )
            self.assertEqual(len(multi_optimization_config.objective.metrics), 2)
            self.assertEqual(
                len(
                    checked_cast(
                        MultiObjectiveOptimizationConfig, multi_optimization_config
                    ).objective_thresholds
                ),
                2,
            )

        with self.subTest(
            "Single-objective optimizations without objective thresholds"
        ):
            single_optimization_config = InstantiationBase.make_optimization_config(
                {"branin": "minimize"},
                objective_thresholds=[],
                outcome_constraints=[],
                status_quo_defined=False,
            )
            self.assertEqual(single_optimization_config.objective.metric.name, "branin")

    def test_single_valued_choice_to_fixed_param_conversion(self) -> None:
        for use_dependents in [True, False]:
            representation: Dict[str, Any] = {
                "name": "test",
                "type": "choice",
                "values": [1.0],
            }
            if use_dependents:
                representation["dependents"] = {1.0: ["foo_or_bar", "bazz"]}
            output = checked_cast(
                FixedParameter, InstantiationBase.parameter_from_json(representation)
            )
            self.assertIsInstance(output, FixedParameter)
            self.assertEqual(output.value, 1.0)
            if use_dependents:
                self.assertEqual(output.dependents, {1.0: ["foo_or_bar", "bazz"]})

    def test_choice_with_is_sorted(self) -> None:
        for sort_values in [True, False, None]:
            representation: Dict[str, Any] = {
                "name": "foo_or_bar",
                "type": "choice",
                "values": ["Foo", "Bar"],
                "sort_values": sort_values,
                "is_ordered": True,
            }
            output = checked_cast(
                ChoiceParameter, InstantiationBase.parameter_from_json(representation)
            )
            self.assertIsInstance(output, ChoiceParameter)
            self.assertEqual(output.is_ordered, True)
            if sort_values is None:
                self.assertIsNone(sort_values)
            else:
                self.assertEqual(output.sort_values, sort_values)

        with self.assertRaisesRegex(ValueError, "Value was not of type <class 'bool'>"):
            representation: Dict[str, Any] = {
                "name": "foo_or_bar",
                "type": "choice",
                "values": ["Foo", "Bar"],
                "sort_values": ["Foo"],
                "is_ordered": True,
            }
            _ = InstantiationBase.parameter_from_json(representation)

    def test_hss(self) -> None:
        parameter_dicts = [
            {
                "name": "root",
                "type": "fixed",
                "value": "HierarchicalSearchSpace",
                "dependents": {"HierarchicalSearchSpace": ["foo_or_bar", "bazz"]},
            },
            {
                "name": "foo_or_bar",
                "type": "choice",
                "values": ["Foo", "Bar"],
                "dependents": {"Foo": ["an_int"], "Bar": ["a_float"]},
            },
            {
                "name": "an_int",
                "type": "choice",
                "values": [1, 2, 3],
                "dependents": None,
            },
            {"name": "a_float", "type": "range", "bounds": [1.0, 1000.0]},
            {
                "name": "bazz",
                "type": "fixed",
                "value": "Bazz",
                "dependents": {"Bazz": ["another_int"]},
            },
            {"name": "another_int", "type": "fixed", "value": "2"},
        ]
        search_space = InstantiationBase.make_search_space(
            # pyre-fixme[6]: For 1st param expected `List[Dict[str, Union[None, Dict[...
            parameters=parameter_dicts,
            parameter_constraints=[],
        )
        self.assertIsInstance(search_space, HierarchicalSearchSpace)
        # pyre-fixme[16]: `SearchSpace` has no attribute `_root`.
        self.assertEqual(search_space._root.name, "root")



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
Module for benchmarking Ax algorithms.

Key terms used:

* Replication: 1 run of an optimization loop; (BenchmarkProblem, BenchmarkMethod) pair.
* Test: multiple replications, ran for statistical significance.
* Full run: multiple tests on many (BenchmarkProblem, BenchmarkMethod) pairs.
* Method: (one of) the algorithm(s) being benchmarked.
* Problem: a synthetic function, a surrogate surface, or an ML model, on which
  to assess the performance of algorithms.

"""

from itertools import product
from logging import Logger
from time import time
from typing import Dict, Iterable, List

import numpy as np

from ax.benchmark.benchmark_method import BenchmarkMethod
from ax.benchmark.benchmark_problem import (
    BenchmarkProblemProtocol,
    BenchmarkProblemWithKnownOptimum,
)
from ax.benchmark.benchmark_result import AggregatedBenchmarkResult, BenchmarkResult
from ax.benchmark.metrics.base import BenchmarkMetricBase, GroundTruthMetricMixin
from ax.core.experiment import Experiment
from ax.core.metric import Metric
from ax.core.objective import MultiObjective, Objective
from ax.core.optimization_config import (
    MultiObjectiveOptimizationConfig,
    OptimizationConfig,
)
from ax.core.outcome_constraint import ObjectiveThreshold, OutcomeConstraint
from ax.core.utils import get_model_times
from ax.service.scheduler import Scheduler
from ax.utils.common.logger import get_logger
from ax.utils.common.random import with_rng_seed
from ax.utils.common.typeutils import checked_cast, not_none

logger: Logger = get_logger(__name__)


def compute_score_trace(
    optimization_trace: np.ndarray,
    num_baseline_trials: int,
    problem: BenchmarkProblemProtocol,
) -> np.ndarray:
    """Computes a score trace from the optimization trace."""

    # Use the first GenerationStep's best found point as baseline. Sometimes (ex. in
    # a timeout) the first GenerationStep will not have not completed and we will not
    # have enough trials; in this case we do not score.
    if (len(optimization_trace) <= num_baseline_trials) or not isinstance(
        problem, BenchmarkProblemWithKnownOptimum
    ):
        return np.full(len(optimization_trace), np.nan)
    optimum = problem.optimal_value
    baseline = optimization_trace[num_baseline_trials - 1]

    score_trace = 100 * (1 - (optimization_trace - optimum) / (baseline - optimum))
    if score_trace.max() > 100:
        logger.info(
            "Observed scores above 100. This indicates that we found a trial that is "
            "better than the optimum. Clipping scores to 100 for now."
        )
    return score_trace.clip(min=0, max=100)


def _create_benchmark_experiment(
    problem: BenchmarkProblemProtocol, method_name: str
) -> Experiment:
    """Creates an empty experiment for the given problem and method.

    If the problem is "noiseless" (i.e. evaluations prior to adding artificial
    noise to the outcomes are determinstic), also adds the respective ground
    truth metrics as tracking metrics for each metric defined on the problem
    (including tracking metrics and the metrics in the OptimizationConfig).

    Args:
        problem: The BenchmarkProblem to test against (can be synthetic or real).
        method_name: Name of the method being tested.

    Returns:
        The Experiment object to be used for benchmarking.
    """
    tracking_metrics = problem.tracking_metrics
    if not problem.is_noiseless and problem.has_ground_truth:
        # Make the ground truth counterparts for each metric defined on the problem,
        # which will be added as tracking metrics on the Experiment object below.
        # In the analysis, a modified OptimziationConfig referencing those metrics
        # will be passed to the `Scheduler.get_trace()` method, which allows to extract
        # the optimziation trace based on the ground truth outcomes (without noise).
        # If the problem is known to be noiseless, this is unneccesary and we can just
        # use the observations made during the optimization loop directly.
        gt_metric_dict = make_ground_truth_metrics(problem=problem)
        tracking_metrics = tracking_metrics + list(gt_metric_dict.values())
    return Experiment(
        name=f"{problem.name}|{method_name}_{int(time())}",
        search_space=problem.search_space,
        optimization_config=problem.optimization_config,
        tracking_metrics=tracking_metrics,  # pyre-ignore [6]: Incompatible
        # parameter type: In call `Experiment.__init__`, for argument
        # `tracking_metrics`, expected `Optional[List[Metric]]` but got
        # `Union[List[Union[BenchmarkMetricBase, Metric]], List[BenchmarkMetricBase]]`.
        runner=problem.runner,
    )


def benchmark_replication(
    problem: BenchmarkProblemProtocol,
    method: BenchmarkMethod,
    seed: int,
) -> BenchmarkResult:
    """Runs one benchmarking replication (equivalent to one optimization loop).

    Args:
        problem: The BenchmarkProblem to test against (can be synthetic or real)
        method: The BenchmarkMethod to test
        seed: The seed to use for this replication.
    """

    experiment = _create_benchmark_experiment(problem=problem, method_name=method.name)

    scheduler = Scheduler(
        experiment=experiment,
        generation_strategy=method.generation_strategy.clone_reset(),
        options=method.scheduler_options,
    )

    with with_rng_seed(seed=seed):
        scheduler.run_n_trials(max_trials=problem.num_trials)

    if not problem.is_noiseless and problem.has_ground_truth:
        # We modify the optimization config so we can use `Scheduler.get_trace()`
        # to use the true (not corrupted by noise) observations that were logged
        # as tracking metrics on the Experiment object. If the problem is known to
        # be noiseless, this is unnecssary and we can just use the observations
        # made during the optimization loop directly.
        analysis_opt_config = make_ground_truth_optimization_config(
            experiment=experiment
        )
    else:
        analysis_opt_config = experiment.optimization_config

    optimization_trace = np.asarray(
        scheduler.get_trace(optimization_config=analysis_opt_config)
    )

    try:
        # Catch any errors that may occur during score computation, such as errors
        # while accessing "steps" in node based generation strategies. The error
        # handling here is intentionally broad. The score computations is not
        # an essential step of the benchmark runs. We do not want to throw away
        # valuable results after the benchmark run completes just because a
        # non-essential step failed.
        num_baseline_trials = scheduler.standard_generation_strategy._steps[
            0
        ].num_trials

        score_trace = compute_score_trace(
            optimization_trace=optimization_trace,
            num_baseline_trials=num_baseline_trials,
            problem=problem,
        )
    except Exception as e:
        logger.warning(
            f"Failed to compute score trace. Returning NaN. Original error message: {e}"
        )
        score_trace = np.full(len(optimization_trace), np.nan)

    fit_time, gen_time = get_model_times(experiment=experiment)

    return BenchmarkResult(
        name=scheduler.experiment.name,
        seed=seed,
        experiment=scheduler.experiment,
        optimization_trace=optimization_trace,
        score_trace=score_trace,
        fit_time=fit_time,
        gen_time=gen_time,
    )


def benchmark_one_method_problem(
    problem: BenchmarkProblemProtocol,
    method: BenchmarkMethod,
    seeds: Iterable[int],
) -> AggregatedBenchmarkResult:
    return AggregatedBenchmarkResult.from_benchmark_results(
        results=[
            benchmark_replication(problem=problem, method=method, seed=seed)
            for seed in seeds
        ]
    )


def benchmark_multiple_problems_methods(
    problems: Iterable[BenchmarkProblemProtocol],
    methods: Iterable[BenchmarkMethod],
    seeds: Iterable[int],
) -> List[AggregatedBenchmarkResult]:
    """
    For each `problem` and `method` in the Cartesian product of `problems` and
    `methods`, run the replication on each seed in `seeds` and get the results
    as an `AggregatedBenchmarkResult`, then return a list of each
    `AggregatedBenchmarkResult`.
    """
    return [
        benchmark_one_method_problem(problem=p, method=m, seeds=seeds)
        for p, m in product(problems, methods)
    ]


def make_ground_truth_metrics(
    problem: BenchmarkProblemProtocol,
    include_tracking_metrics: bool = True,
) -> Dict[str, Metric]:
    """Makes a ground truth version for each metric defined on the problem.

    Args:
        problem: The BenchmarkProblem to test against (can be synthetic or real).
        include_tracking_metrics: Whether or not to include tracking metrics.

    Returns:
        A dict mapping (original) metric names to their respective ground truth metric.
    """
    if not problem.has_ground_truth:
        raise ValueError(
            "Cannot create ground truth metrics for problems that "
            "do not have a ground truth."
        )
    metrics: List[BenchmarkMetricBase] = [
        checked_cast(BenchmarkMetricBase, metric)
        for metric in problem.optimization_config.metrics.values()
    ]
    if include_tracking_metrics:
        metrics = metrics + problem.tracking_metrics
    return {metric.name: metric.make_ground_truth_metric() for metric in metrics}


def make_ground_truth_optimization_config(
    experiment: Experiment,
) -> OptimizationConfig:
    """Makes a clone of the OptimizationConfig on the experiment in which each metric
    is replaced by its respective "ground truth" counterpart, which has been added to
    the experiment's tracking metrics in `_create_benchmark_experiment` and which
    returns the ground truth (i.e., uncorrupted by noise) observations.
    """
    optimization_config = not_none(experiment.optimization_config)

    if optimization_config.risk_measure is not None:
        raise NotImplementedError("Support for risk measures is not yet implemented.")

    # dict for caching metric lookup
    gt_metric_dict: Dict[str, BenchmarkMetricBase] = {}

    def get_gt_metric(metric: Metric) -> BenchmarkMetricBase:
        """Look up corresponding ground truth metric of the experiment. Will error
        out if no corresponding ground truth metric exists."""
        if not isinstance(metric, BenchmarkMetricBase):
            raise ValueError(
                "Only BenchmarkMetricBase metrics are supported for ground truth "
                f"metrics. Got {type(metric)}."
            )

        if metric.name in gt_metric_dict:
            return gt_metric_dict[metric.name]

        for tracking_metric in experiment.tracking_metrics:
            if getattr(tracking_metric, "is_ground_truth", False):
                # TODO: Figure out if there is a better way to match the ground truth
                # metric and the original metric.
                ground_truth_name = tracking_metric.name
                orig_name = checked_cast(
                    GroundTruthMetricMixin, tracking_metric
                ).get_original_name(ground_truth_name)
                if orig_name == metric.name:
                    tracking_metric = checked_cast(BenchmarkMetricBase, tracking_metric)
                    gt_metric_dict[metric.name] = tracking_metric
                    return tracking_metric
        raise ValueError(f"Ground truth metric for metric {metric.name} not found!")

    # convert outcome constraints
    if optimization_config.outcome_constraints is not None:
        gt_outcome_constraints = [
            OutcomeConstraint(
                metric=get_gt_metric(oc.metric),
                op=oc.op,
                bound=oc.bound,
                relative=oc.relative,
            )
            for oc in optimization_config.outcome_constraints
        ]
    else:
        gt_outcome_constraints = None

    # we need to distinguish MOO and non-MOO problems
    if not optimization_config.is_moo_problem:
        gt_objective = Objective(
            metric=get_gt_metric(optimization_config.objective.metric)
        )

        return OptimizationConfig(
            objective=gt_objective, outcome_constraints=gt_outcome_constraints
        )

    gt_objective = MultiObjective(
        metrics=[
            get_gt_metric(metric) for metric in optimization_config.objective.metrics
        ]
    )
    # there may be objective thresholds to also convert
    objective_thresholds = checked_cast(
        MultiObjectiveOptimizationConfig, optimization_config
    ).objective_thresholds
    if objective_thresholds is not None:
        gt_objective_thresholds = [
            ObjectiveThreshold(
                metric=get_gt_metric(ot.metric),
                bound=ot.bound,
                relative=ot.relative,
                op=ot.op,
            )
            for ot in objective_thresholds
        ]
    else:
        gt_objective_thresholds = None

    return MultiObjectiveOptimizationConfig(
        objective=gt_objective,
        outcome_constraints=gt_outcome_constraints,
        objective_thresholds=gt_objective_thresholds,
    )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from inspect import Parameter, signature

from logging import Logger
from typing import Any, Callable, Dict, Iterable, List, Optional

from ax.utils.common.logger import get_logger
from ax.utils.common.typeutils_nonnative import version_safe_check_type

logger: Logger = get_logger(__name__)

TKwargs = Dict[str, Any]


def consolidate_kwargs(
    kwargs_iterable: Iterable[Optional[Dict[str, Any]]], keywords: Iterable[str]
) -> Dict[str, Any]:
    """Combine an iterable of kwargs into a single dict of kwargs, where kwargs
    by duplicate keys that appear later in the iterable get priority over the
    ones that appear earlier and only kwargs referenced in keywords will be
    used. This allows to combine somewhat redundant sets of kwargs, where a
    user-set kwarg, for instance, needs to override a default kwarg.

    >>> consolidate_kwargs(
    ...     kwargs_iterable=[{'a': 1, 'b': 2}, {'b': 3, 'c': 4, 'd': 5}],
    ...     keywords=['a', 'b', 'd']
    ... )
    {'a': 1, 'b': 3, 'd': 5}
    """
    all_kwargs = {}
    for kwargs in kwargs_iterable:
        if kwargs is not None:
            all_kwargs.update({kw: p for kw, p in kwargs.items() if kw in keywords})
    return all_kwargs


def get_function_argument_names(
    # pyre-fixme[24]: Generic type `Callable` expects 2 type parameters.
    function: Callable,
    omit: Optional[List[str]] = None,
) -> List[str]:
    """Extract parameter names from function signature."""
    omit = omit or []
    return [p for p in signature(function).parameters.keys() if p not in omit]


# pyre-fixme[24]: Generic type `Callable` expects 2 type parameters.
def get_function_default_arguments(function: Callable) -> Dict[str, Any]:
    """Extract default arguments from function signature."""
    params = signature(function).parameters
    return {
        kw: p.default for kw, p in params.items() if p.default is not Parameter.empty
    }


# pyre-fixme[24]: Generic type `Callable` expects 2 type parameters.
def validate_kwarg_typing(typed_callables: List[Callable], **kwargs: Any) -> None:
    """Check if keywords in kwargs exist in any of the typed_callables and
    if the type of each keyword value matches the type of corresponding arg in one of
    the callables

    Note: this function expects the typed callables to have unique keywords for
    the arguments and will raise an error if repeat keywords are found.
    """
    checked_kwargs = set()
    for typed_callable in typed_callables:
        params = signature(typed_callable).parameters
        for kw, param in params.items():
            if kw in kwargs:
                if kw in checked_kwargs:
                    logger.debug(
                        f"`{typed_callables}` have duplicate keyword argument: {kw}."
                    )
                else:
                    checked_kwargs.add(kw)
                    kw_val = kwargs.get(kw)
                    # if the keyword is a callable, we only do shallow checks
                    if not (callable(kw_val) and callable(param.annotation)):
                        try:
                            version_safe_check_type(kw, kw_val, param.annotation)
                        except TypeError:
                            message = (
                                f"`{typed_callable}` expected argument `{kw}` to be of"
                                f" type {param.annotation}. Got {kw_val}"
                                f" (type: {type(kw_val)})."
                            )
                            logger.warning(message)

    # check if kwargs contains keywords not exist in any callables
    extra_keywords = [kw for kw in kwargs.keys() if kw not in checked_kwargs]
    if len(extra_keywords) != 0:
        raise ValueError(
            f"Arguments {extra_keywords} are not expected by any of {typed_callables}."
        )


# pyre-fixme[24]: Generic type `Callable` expects 2 type parameters.
def warn_on_kwargs(callable_with_kwargs: Callable, **kwargs: Any) -> None:
    """Log a warning when a decoder function receives unexpected kwargs.

    NOTE: This mainly caters to the use case where an older version of Ax is
    used to decode objects, serialized to JSON by a newer version of Ax (and
    therefore potentially containing new fields). In that case, the decoding
    function should not fail when encountering those additional fields, but
    rather just ignore them and log a warning using this function.
    """
    if kwargs:
        logger.warning(
            "Found unexpected kwargs: %s while calling %s "
            "from JSON. These kwargs will be ignored.",
            kwargs,
            callable_with_kwargs,
        )


# pyre-fixme[3]: Return annotation cannot be `Any`.
# pyre-fixme[24]: Generic type `Callable` expects 2 type parameters.
def filter_kwargs(function: Callable, **kwargs: Any) -> Any:
    """Filter out kwargs that are not applicable for a given function.
    Return a copy of given kwargs dict with only the required kwargs."""
    return {k: v for k, v in kwargs.items() if k in signature(function).parameters}



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import numpy as np
from ax.metrics.noisy_function import NoisyFunctionMetric


class L2NormMetric(NoisyFunctionMetric):
    def f(self, x: np.ndarray) -> float:
        return np.sqrt((x**2).sum())



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from unittest.mock import MagicMock

import torch
from ax.benchmark.problems.surrogate import SurrogateRunner
from ax.core.parameter import ParameterType, RangeParameter
from ax.core.search_space import SearchSpace
from ax.utils.common.testutils import TestCase


class TestSurrogateRunner(TestCase):
    def test_surrogate_runner(self) -> None:
        # Construct a search space with log-scale parameters.
        search_space = SearchSpace(
            parameters=[
                RangeParameter("x", ParameterType.FLOAT, 0.0, 5.0),
                RangeParameter("y", ParameterType.FLOAT, 1.0, 10.0, log_scale=True),
                RangeParameter("z", ParameterType.INT, 1.0, 5.0, log_scale=True),
            ]
        )
        for noise_std in (0.0, 0.1, {"dummy_metric": 0.2}):
            with self.subTest(noise_std=noise_std):
                surrogate = MagicMock()
                mock_mean = torch.tensor([[0.1234]], dtype=torch.double)
                surrogate.predict = MagicMock(return_value=(mock_mean, 0))
                surrogate.device = torch.device("cpu")
                surrogate.dtype = torch.double
                runner = SurrogateRunner(
                    name="test runner",
                    surrogate=surrogate,
                    datasets=[],
                    search_space=search_space,
                    outcome_names=["dummy_metric"],
                    noise_stds=noise_std,
                )
                self.assertEqual(runner.name, "test runner")
                self.assertIs(runner.surrogate, surrogate)
                self.assertEqual(runner.outcome_names, ["dummy_metric"])
                self.assertEqual(runner.noise_stds, noise_std)
